/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs;

import java.io.File;
import java.io.PrintStream;
import java.net.URL;

import javax.swing.UIManager;

import jfs.conf.JFSConfig;
import jfs.conf.JFSConst;
import jfs.conf.JFSDirectoryPair;
import jfs.conf.JFSFilter;
import jfs.conf.JFSLog;
import jfs.conf.JFSSettings;
import jfs.conf.JFSText;
import jfs.gui.JFSMainView;
import jfs.server.JFSServerFactory;
import jfs.shell.JFSShell;
import jfs.sync.JFSTask;

/**
 * JFileSync is an application for synchronizing pairs of directories. This
 * class is the main class of the application. It processes all command line
 * options and starts (1) a JFS server, (2) the Java Swing-based graphical user
 * interface, or (3) a command line shell.
 * 
 * @see JFSServerFactory#startCmdLineServer()
 * @see jfs.gui.JFSMainView
 * @see JFSShell#startShell(boolean)
 * @author Jens Heidrich
 * @version $Id: JFileSync.java,v 1.31 2005/10/04 21:48:47 heidrich Exp $
 */
public class JFileSync {

	/**
	 * Determines the home directory, where the JAR file respectively the class
	 * files are located.
	 */
	public static String getHome() {
		try {
			URL packageUrl = ClassLoader.getSystemResource("jfs");
			File jfsLibDir = null;

			// JFS started from JAR:
			if (packageUrl.getProtocol().equals("jar")) {
				URL jarUrl = new URL(packageUrl.getFile());
				File jfsFile = new File(jarUrl.toURI());
				jfsLibDir = jfsFile.getParentFile().getParentFile();
			}

			// JFS started from classes directory:
			if (packageUrl.getProtocol().equals("file")) {
				File jfsFile = new File(packageUrl.toURI());
				jfsLibDir = jfsFile.getParentFile();
			}

			return jfsLibDir.getPath();
		} catch (Exception e) {
			return ".";
		}
	}

	/**
	 * Start of the application.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		// Get settings for the first time in order to load stored
		// settings before doing any actions:
		JFSSettings s = JFSSettings.getInstance();

		// Get translation and configuration object:
		JFSText t = JFSText.getInstance();
		JFSConfig config = JFSConfig.getInstance();
		PrintStream p = JFSLog.getOut().getStream();

		boolean nogui = false;
		boolean quiet = false;
		boolean launchServer = false;
		boolean loadDefaultFile = true;
		int i = 0;

		// Handle command line arguments:
		try {
			while (i < args.length) {
				if (args[i].equals("-config")) {
					i++;
					config.load(new File(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-sync")) {
					i++;
					config.setSyncMode(Byte.parseByte(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-view")) {
					i++;
					config.setView(Byte.parseByte(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-dir")) {
					i++;
					config.addDirectoryPair(new JFSDirectoryPair(args[i],
							args[i + 1]));
					i++;
					loadDefaultFile = false;
				} else if (args[i].equals("-granularity")) {
					i++;
					config.setGranularity(Integer.parseInt(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-buffersize")) {
					i++;
					config.setBufferSize(Integer.parseInt(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-test")) {
					config.setTestRun(true);
					loadDefaultFile = false;
				} else if (args[i].equals("-nohistory")) {
					config.setStoreHistory(false);
					loadDefaultFile = false;
				} else if (args[i].equals("-include")) {
					i++;
					config.addInclude(new JFSFilter(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-exclude")) {
					i++;
					config.addExclude(new JFSFilter(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-server")) {
					launchServer = true;
					if (((i + 1) < args.length) && !args[i + 1].startsWith("-")) {
						i++;
						config.setServerBase(args[i]);
					}
					loadDefaultFile = false;
				} else if (args[i].equals("-port")) {
					i++;
					config.setServerPort(Integer.parseInt(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-passphrase")) {
					i++;
					config.setServerPassPhrase(args[i]);
					loadDefaultFile = false;
				} else if (args[i].equals("-timeout")) {
					i++;
					config.setServerTimeout(Integer.parseInt(args[i]));
					loadDefaultFile = false;
				} else if (args[i].equals("-shutdown")) {
					config.setServerShutDown(true);
					loadDefaultFile = false;
				} else if (args[i].equals("-nogui")) {
					nogui = true;
				} else if (args[i].equals("-quiet")) {
					quiet = true;
				} else if (args[i].equals("-laf")) {
					if (((i + 1) >= args.length) || args[i + 1].startsWith("-")) {
						p.println(t.get("CMD_LAF"));

						UIManager.LookAndFeelInfo[] laf = UIManager
								.getInstalledLookAndFeels();

						for (int j = 0; j < laf.length; j++) {
							p.println(j + 1 + ". " + laf[j].getName() + " <"
									+ laf[j].getClassName() + ">");
						}

						System.exit(0);
					} else {
						i++;
						s.setLaf(args[i]);
					}
				} else if (args[i].equals("-help") || args[i].equals("-?")) {
					JFSShell.printURL(JFSConst.getInstance().getResourceUrl(
							"HELP_CMD_LINE"));
					System.exit(0);
				} else if (args[i].equals("-version")) {
					p.println(t.get("JFILESYNC") + " "
							+ JFSConst.getInstance().getString("VERSION"));
					p.println(t.get("INFO_COPYRIGHT"));
					p.println(t.get("INFO_AUTHOR"));
					p.println(t.get("INFO_JFS_HOME") + " " + getHome());
					System.exit(0);
				} else if (args.length == 1) {
					config.load(new File(args[i]));
					loadDefaultFile = false;
				} else {
					JFSShell.printURL(JFSConst.getInstance().getResourceUrl(
							"HELP_CMD_LINE"));
					System.exit(0);
				}

				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			JFSShell.printURL(JFSConst.getInstance().getResourceUrl(
					"HELP_CMD_LINE"));
			System.exit(0);
		} catch (NumberFormatException e) {
			JFSLog.getErr().getStream().println(t.get("ERROR_NUMBER_FORMAT"));
			System.exit(0);
		}

		// Start server and exit afterwards if server should be launched:
		if (launchServer) {
			JFSServerFactory.getInstance().startCmdLineServer();
		} else {
			// Initialize new task object:
			JFSTask task = JFSTask.getInstance();
			config.attach(task);

			if (!nogui) {
				// Start GUI:
				p.println(t.get("CMD_START_GUI"));

				// Determine whether the last configuration when (stored when
				// the program was exited should be loaded at GUI startup:
				new JFSMainView(loadDefaultFile);
			} else {
				JFSShell.startShell(quiet);
			}
		}
	}
}