/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.conf;

import java.util.regex.PatternSyntaxException;

/**
 * This class specifies a single filter used to identify files that should be
 * excluded from synchronization. Files matching the specified filter will not
 * be added to the synchronization list.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSFilter.java,v 1.4 2005/09/28 20:33:11 heidrich Exp $
 */
public class JFSFilter implements Cloneable {

	/** Determines whether the filter should be applied or not. */
	private boolean isActive;

	/** The regular expression representing the filter. */
	private String filter;

	/**
	 * Creates a new filter.
	 * 
	 * @param filter
	 *            The filter to use.
	 */
	public JFSFilter(String filter) {
		isActive = true;
		this.filter = filter;
	}

	/**
	 * @return Returns the regular expression of the filter.
	 */
	public String getFilter() {
		return filter;
	}

	/**
	 * @param filter
	 *            Sets the regular expression of the filter.
	 */
	public void setFilter(String filter) {
		this.filter = filter;
	}

	/**
	 * @return Returns the whether to filter is active.
	 */
	public boolean isActive() {
		return isActive;
	}

	/**
	 * @param isActive
	 *            Activates or deactivates the filter.
	 */
	public void setActive(boolean isActive) {
		this.isActive = isActive;
	}

	/**
	 * Tests whether a given string machtes the filter. If the filter is not
	 * active false is returned. If the filter is active, the regular expression
	 * of the filter is tested against a given string.
	 * 
	 * @param str
	 *            The string to test.
	 * @return Returns true if the filter is active and the given string matches
	 *         the regular expression of the filter.
	 */
	public boolean matches(String str) {
		try {
			return isActive && str.matches(filter);
		} catch (PatternSyntaxException e) {
			return false;
		}
	}

	/**
	 * @see Object#clone()
	 */
	public JFSFilter clone() {
		JFSFilter clone = new JFSFilter(filter);
		clone.setActive(isActive);
		return clone;
	}

	/**
	 * @see Object#equals(Object)
	 */
	public boolean equals(JFSFilter o) {
		return filter.equals(o.getFilter()) && isActive == o.isActive();
	}
}