/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.conf;

import java.io.File;
import java.text.DateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Vector;

import jfs.sync.JFSElement;
import jfs.sync.JFSElement.ElementState;

/**
 * Handles a history of synchronized files for a certain directory pair.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSHistory.java,v 1.8 2005/10/05 22:06:45 heidrich Exp $
 */
public abstract class JFSHistory implements Comparable<JFSHistory> {
	/** The directory pair the history is created for. */
	private JFSDirectoryPair pair = null;

	/** The date when the history was created or updated. */
	private long date = -1;

	/** The items of the history. */
	protected Vector<JFSHistoryItem> history = new Vector<JFSHistoryItem>();

	/** The directory items of the history for fast access. */
	protected HashMap<String, JFSHistoryItem> directories = new HashMap<String, JFSHistoryItem>();

	/** The file items of the history for fast access. */
	protected HashMap<String, JFSHistoryItem> files = new HashMap<String, JFSHistoryItem>();

	/** The assigned file name the history is stored in and loaded from. */
	private String fileName = null;

	/** Determines whether the history was already loaded from file. */
	private boolean isLoaded = false;

	/**
	 * Clears all history elements apart from the assigned directory pair.
	 */
	public void clear() {
		date = -1;
		history.clear();
		directories.clear();
		files.clear();
		fileName = null;
		isLoaded = false;
	}

	/**
	 * Loads a history. If the history is already loaded true is returned.
	 * 
	 * @return True if and only if loading was successful.
	 */
	public boolean load() {
		if (isLoaded)
			return true;

		// Return false, if no pair or no file was assigned:
		if (pair == null || fileName == null)
			return false;

		File file = new File(JFSConst.HOME_DIR + File.separatorChar + fileName);
		isLoaded = load(file);

		return isLoaded;
	}

	/**
	 * Loads a history.
	 * 
	 * @param file
	 *            The history to load.
	 * @return True if and only if loading did not fail.
	 */
	protected abstract boolean load(File file);

	/**
	 * Stores a history. If no pair is assigned, false is returned. If no file
	 * is assigned a new file is created in the JFS configuration directory. If
	 * the history is not part if the history manager, it is added to the
	 * manager.
	 * 
	 * @return True if and only if storing was successful.
	 */
	public boolean store() {
		// Return false, if no pair is assigned:
		if (pair == null)
			return false;

		// Create new unique file name, if no file is assigned:
		File file = null;
		if (fileName == null) {
			long time = System.currentTimeMillis();
			do {
				fileName = JFSConst.HISTORY_FILE_PREFIX + time + ".xml";
				file = new File(JFSConst.HOME_DIR + File.separatorChar
						+ fileName);
				time += 1;
			} while (file.exists());
		} else {
			file = new File(JFSConst.HOME_DIR + File.separatorChar + fileName);
		}

		// Check that history is in manager:
		JFSHistoryManager hm = JFSHistoryManager.getInstance();
		if (!hm.getHistories().contains(this))
			hm.addHistory(this);
		hm.sortHistories();

		boolean success = store(file);

		// Store settings when history is saved successfully:
		if (success)
			JFSSettings.getInstance().store();

		return success;
	}

	/**
	 * Stores a history.
	 * 
	 * @param file
	 *            The history to store.
	 * @return True if and only if storing did not fail.
	 */
	protected abstract boolean store(File file);

	/**
	 * Returns the history item for a certain JFS element or null if no item was
	 * found.
	 * 
	 * @param element
	 *            The JFS element to search a history item for.
	 * @return The corresponding history item.
	 */
	public final JFSHistoryItem getHistory(JFSElement element) {
		if (element.isDirectory()) {
			return directories.get(element.getRelativePath());
		} else {
			return files.get(element.getRelativePath());
		}
	}

	/**
	 * Returns the date when the history was created/updated.
	 * 
	 * @return The date.
	 */
	public long getDate() {
		return date;
	}

	/**
	 * Returns a date string for the last modification of the file.
	 * 
	 * @return Time of last modification of the file as a date string.
	 */
	public final String getDateAsString() {
		if (getDate() != -1) {
			DateFormat df = DateFormat.getDateTimeInstance(DateFormat.SHORT,
					DateFormat.SHORT);
			return df.format(new Date(getDate()));
		} else {
			return "-";
		}
	}

	/**
	 * Sets the date when the history was created/updated.
	 * 
	 * @param date
	 *            The date to set.
	 */
	public void setDate(long date) {
		this.date = date;
	}

	/**
	 * Returns the assigned directory pair.
	 * 
	 * @return The directory pair.
	 */
	public JFSDirectoryPair getPair() {
		return pair;
	}

	/**
	 * Sets the assigned directory pair.
	 * 
	 * @param pair
	 *            The directory pair to set.
	 */
	public void setPair(JFSDirectoryPair pair) {
		this.pair = pair;
	}

	/**
	 * Returns the file to load the history from and to store the history to.
	 * 
	 * @return The file name.
	 */
	public String getFileName() {
		return fileName;
	}

	/**
	 * Sets the file to load the history from and to store the history to.
	 * 
	 * @param fileName
	 *            The file name to set.
	 */
	public void setFileName(String fileName) {
		this.fileName = fileName;
	}

	/**
	 * Updates the history with a comparison table. If the element's files are
	 * equal (identified by their equal time stamp), an existing history item
	 * has to be updated or a new one has to be created. If a history item was
	 * found (and updated) or a new one was created, it is added to the new
	 * history. The relative path is stored as well as the last modified date
	 * and the length. That means, if the JFS element's files are not equal, the
	 * previous history item is kept as is. The latter can be caused by an
	 * interrupted or failed synchronization process or if JFS elements are
	 * deactivated before synchronization.
	 * 
	 * @param comparisonTable
	 *            The comparison table to be used for updating the history.
	 */
	public void update(Vector<JFSElement> comparisonTable) {
		Vector<JFSHistoryItem> newHistory = new Vector<JFSHistoryItem>();
		HashMap<String, JFSHistoryItem> newDirectories = new HashMap<String, JFSHistoryItem>();
		HashMap<String, JFSHistoryItem> newFiles = new HashMap<String, JFSHistoryItem>();

		setDate(System.currentTimeMillis());

		for (JFSElement element : comparisonTable) {
			JFSHistoryItem item = getHistory(element);

			// If the element's files are equal, an existing item has to be
			// updated or a new one has to be created:
			if (element.getState() == ElementState.EQUAL) {
				// Use the source file (which exists if both files are equal)
				// to construct the new history item:
				if (item == null) {
					item = new JFSHistoryItem(element.getRelativePath());
					item.setDirectory(element.isDirectory());
				}
				item.setLastModified(element.getSrcFile().getLastModified());
				item.setLength(element.getSrcFile().getLength());
			}

			// If a history item was found or a new one was created, add this
			// to the new history:
			if (item != null) {
				newHistory.add(item);
				if (item.isDirectory()) {
					newDirectories.put(item.getRelativePath(), item);
				} else {
					newFiles.put(item.getRelativePath(), item);
				}
			}
		}

		// Replace old history:
		history = newHistory;
		directories = newDirectories;
		files = newFiles;
	}

	/**
	 * @see Comparable#compareTo(Object)
	 */
	public int compareTo(JFSHistory h) {
		if (getDate() > h.getDate()) {
			return 1;
		} else if (getDate() < h.getDate()) {
			return -1;
		} else {
			return 0;
		}
	}
}