/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.conf;

import java.io.File;
import java.util.Collections;
import java.util.Vector;

/**
 * Manager all histories of synchronized directory pairs.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSHistoryManager.java,v 1.4 2005/09/28 20:33:11 heidrich Exp $
 */
public class JFSHistoryManager {
	/** Stores the only instance of the class. */
	private static JFSHistoryManager instance = null;

	/** The stored histories. */
	private Vector<JFSHistory> histories = new Vector<JFSHistory>();

	/**
	 * Constructs a the only history manager.
	 */
	private JFSHistoryManager() {
	}

	/**
	 * Returns the reference of the only instance.
	 * 
	 * @return The only instance.
	 */
	public final static JFSHistoryManager getInstance() {
		if (instance == null)
			instance = new JFSHistoryManager();

		return instance;
	}

	/**
	 * Sorts all histories stored.
	 */
	public final void sortHistories() {
		Collections.sort(histories);
	}

	/**
	 * Returns the set of histories.
	 * 
	 * @return The histories.
	 */
	public final Vector<JFSHistory> getHistories() {
		return histories;
	}

	/**
	 * Returns the history for a special directory pair. If no history is found,
	 * a new one is created and added to the list of histories.
	 * 
	 * @param pair
	 *            The directory pair to search the history for.
	 * @return The history.
	 */
	public final JFSHistory getHistory(JFSDirectoryPair pair) {
		for (JFSHistory h : histories)
			if (h.getPair().equals(pair))
				return h;
		JFSHistory h = new JFSHistoryXML();
		h.setPair(pair);
		histories.add(h);
		return h;
	}

	/**
	 * Adds a special history.
	 * 
	 * @param history
	 *            The history to add.
	 */
	public final void addHistory(JFSHistory history) {
		histories.add(history);
	}

	/**
	 * Deletes the history for a special history; that is, deletes the file, the
	 * history is stored to and removes the history from the list of histories.
	 * 
	 * @param history
	 *            The history to delete.
	 */
	public final void deleteHistory(JFSHistory history) {
		deleteHistoryFile(history);
		history.clear();
		histories.remove(history);
	}

	/**
	 * Deletes all histories.
	 * 
	 * @see #deleteHistory(JFSHistory)
	 */
	public final void deleteAll() {
		for (JFSHistory h : histories) {
			deleteHistoryFile(h);
			h.clear();
		}
		histories.clear();
	}

	/**
	 * Deletes the history file for a special history.
	 * 
	 * @param history
	 *            The history to delete the corresponding file for.
	 */
	private final void deleteHistoryFile(JFSHistory history) {
		File file = new File(JFSConst.HOME_DIR + File.separatorChar
				+ history.getFileName());
		if (file != null && file.exists())
			file.delete();
	}

	/**
	 * Cleans the JFS configuration directory by cleaning all history files that
	 * are not referenced in the history manager.
	 */
	public final void cleanHistories() {
		Vector<String> historyFiles = new Vector<String>();
		for (JFSHistory h : getHistories()) {
			historyFiles.add(h.getFileName());
		}

		File jfsDir = new File(JFSConst.HOME_DIR);
		for (File f : jfsDir.listFiles()) {
			String name = f.getName();
			if (name.startsWith(JFSConst.HISTORY_FILE_PREFIX)
					&& name.endsWith(".xml") && !historyFiles.contains(name)) {
				f.delete();
			}
		}
	}
}