/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;

import jfs.conf.JFSConst;
import jfs.conf.JFSText;
import jfs.conf.JFSSyncMode.SyncAction;
import jfs.sync.JFSElement;
import jfs.sync.JFSFile;
import jfs.sync.JFSTask;

/**
 * This class is responsible for displaying the comparison table.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSComparisonTable.java,v 1.23 2005/10/04 19:32:37 heidrich Exp $
 */
public class JFSComparisonTable extends AbstractTableModel implements
		TableCellRenderer, MouseListener, ActionListener {
	/** The UID. */
	private static final long serialVersionUID = 45L;

	/** Color definition for files that have to be copied from source. */
	private static final Color COPY_SRC = new Color(130, 255, 130);

	/** Color definition for files that have to be copied from target. */
	private static final Color COPY_TGT = new Color(130, 130, 255);

	/** Color definition for files that have to be deleted. */
	private static final Color DELETE = new Color(255, 130, 130);

	/** Color definition for a row that contains directories. */
	private static final Color DIRECTORY = new Color(255, 255, 200);

	/** Color definition for a warning. */
	private static final Color WARNING = new Color(255, 0, 0);

	/** Color definition for a warning. */
	private static final Color WARNING_SELECTED = new Color(255, 220, 220);

	/** The corresponding JTable. */
	private final JTable localTable;

	/** An adapted JLabel object for our cell renderer component. */
	protected JLabel cell;

	/** The object with information that should be displayed by the table. */
	private JFSTask task;

	/** Stores the names for the icons used for displaying actions. */
	private HashMap<SyncAction, String> actionIconNames = new HashMap<SyncAction, String>();

	/** Stores the icons used for displaying the action for a JFS element. */
	private HashMap<SyncAction, JLabel> actionIcons = new HashMap<SyncAction, JLabel>();

	/** Stores the icons used for displaying the action for a JFS element. */
	private HashMap<SyncAction, Color> actionBackgrounds = new HashMap<SyncAction, Color>();

	/** Stores the current selection of JFS elements in the table. */
	private Vector<JFSElement> currentSelection = new Vector<JFSElement>();

	/**
	 * The default constructor just performs some initialization work.
	 * 
	 * @param task
	 *            The task object.
	 */
	public JFSComparisonTable(JFSTask task) {
		this.task = task;

		// Create column model:
		JFSConst jfsConst = JFSConst.getInstance();
		JFSText t = JFSText.getInstance();
		DefaultTableColumnModel cm = new DefaultTableColumnModel();

		TableColumn srcFile = new TableColumn(0, 205);
		srcFile.setHeaderValue(t.get("COMP_TABLE_SRC"));
		cm.addColumn(srcFile);

		TableColumn srcDate = new TableColumn(1, 120);
		srcDate.setHeaderValue(t.get("COMP_TABLE_DATE"));
		cm.addColumn(srcDate);

		TableColumn srcSize = new TableColumn(2, 60);
		srcSize.setHeaderValue(t.get("COMP_TABLE_SIZE"));
		cm.addColumn(srcSize);

		TableColumn comp = new TableColumn(3, 30);
		comp.setHeaderValue(t.get("COMP_TABLE_ACTION"));
		cm.addColumn(comp);

		TableColumn tgtFile = new TableColumn(4, 205);
		tgtFile.setHeaderValue(t.get("COMP_TABLE_TGT"));
		cm.addColumn(tgtFile);

		TableColumn tgtDate = new TableColumn(5, 120);
		tgtDate.setHeaderValue(t.get("COMP_TABLE_DATE"));
		cm.addColumn(tgtDate);

		TableColumn tgtSize = new TableColumn(6, 60);
		tgtSize.setHeaderValue(t.get("COMP_TABLE_SIZE"));
		cm.addColumn(tgtSize);

		// Create table and cell labels:
		localTable = new JTable(this, cm);
		localTable.setDefaultRenderer(String.class, this);
		localTable.getTableHeader().setReorderingAllowed(false);
		cell = new JLabel();
		cell.setOpaque(true);
		cell.setFont(localTable.getFont());

		// Set action icons:
		actionIconNames.put(SyncAction.NOP, "ICON_IDLE");
		actionIconNames.put(SyncAction.COPY_SRC, "ICON_COPY_SRC");
		actionIconNames.put(SyncAction.COPY_TGT, "ICON_COPY_TGT");
		actionIconNames.put(SyncAction.DELETE_SRC, "ICON_DELETE");
		actionIconNames.put(SyncAction.DELETE_TGT, "ICON_DELETE");
		actionIconNames.put(SyncAction.DELETE_SRC_AND_TGT, "ICON_DELETE");
		actionIconNames
				.put(SyncAction.ASK_LENGTH_INCONSISTENT, "ICON_QUESTION");
		actionIconNames.put(SyncAction.ASK_FILES_GT_HISTORY, "ICON_QUESTION");
		actionIconNames.put(SyncAction.ASK_FILES_NOT_IN_HISTORY,
				"ICON_QUESTION");

		for (SyncAction a : actionIconNames.keySet()) {
			JLabel icon = new JLabel(new ImageIcon(jfsConst
					.getIconUrl(actionIconNames.get(a))));
			icon.setOpaque(true);
			icon.setHorizontalAlignment(JLabel.CENTER);
			actionIcons.put(a, icon);
		}

		// Set action backgrounds:
		Color background = localTable.getBackground();
		actionBackgrounds.put(SyncAction.NOP, background);
		actionBackgrounds.put(SyncAction.COPY_SRC, COPY_SRC);
		actionBackgrounds.put(SyncAction.COPY_TGT, COPY_TGT);
		actionBackgrounds.put(SyncAction.DELETE_SRC, DELETE);
		actionBackgrounds.put(SyncAction.DELETE_TGT, DELETE);
		actionBackgrounds.put(SyncAction.DELETE_SRC_AND_TGT, DELETE);
		actionBackgrounds.put(SyncAction.ASK_LENGTH_INCONSISTENT, background);
		actionBackgrounds.put(SyncAction.ASK_FILES_GT_HISTORY, background);
		actionBackgrounds.put(SyncAction.ASK_FILES_NOT_IN_HISTORY, background);

		// Register mouse listener:
		localTable.addMouseListener(this);
	}

	/**
	 * @see TableModel#getRowCount()
	 */
	public final int getRowCount() {
		return task.getViewTable().size();
	}

	/**
	 * @see TableModel#getColumnCount()
	 */
	public final int getColumnCount() {
		return 7;
	}

	/**
	 * @see TableModel#isCellEditable(int, int)
	 */
	public boolean isCellEditable(int rowIndex, int columnIndex) {
		return false;
	}

	/**
	 * @see TableModel#getValueAt(int, int)
	 */
	public final Object getValueAt(int row, int column) {
		if (row >= task.getViewTable().size())
			return null;

		String entry = "";
		JFSElement element = task.getViewTable().get(row);
		JFSFile file = null;

		if (column >= 0 && column <= 2)
			file = element.getSrcFile();
		else if (column >= 4 && column <= 6)
			file = element.getTgtFile();

		if (file != null)
			switch (column) {
			case 0:
			case 4:
				if (element.isDirectory())
					entry = file.getPath();
				else
					entry = file.getName();

				break;
			case 1:
			case 5:
				entry = file.getLastModifiedDate();
				break;
			case 2:
			case 6:
				entry = file.getLengthString();
				break;
			}

		if (column == 3)
			entry = String.valueOf(element.getAction());

		return entry;
	}

	/**
	 * @see TableModel#getColumnClass(int)
	 */
	public final Class<?> getColumnClass(int column) {
		return String.class;
	}

	/**
	 * @return Returns the JTable object.
	 */
	public final JTable getJTable() {
		return localTable;
	}

	/**
	 * @see TableCellRenderer#getTableCellRendererComponent(JTable, Object,
	 *      boolean, boolean, int, int)
	 */
	public Component getTableCellRendererComponent(JTable table, Object value,
			boolean isSelected, boolean hasFocus, int row, int column) {
		if (row >= task.getViewTable().size() || row < 0)
			return null;

		// Determine corresponding element of the comparison table:
		JFSElement element = task.getViewTable().get(row);

		// Get the right component first and set alignment:
		JComponent component;
		if (column == 3) {
			component = actionIcons.get(element.getAction());
			if (component == null)
				component = actionIcons.get(SyncAction.NOP);
		} else {
			component = cell;
			cell.setText(String.valueOf(value));
			if (column == 2 || column == 6) {
				cell.setHorizontalAlignment(JLabel.RIGHT);
			} else {
				cell.setHorizontalAlignment(JLabel.LEFT);
			}
		}

		// Set tool tip text for component:
		if (column == 3) {
			component.setToolTipText(JFSText.getInstance().get(
					String.valueOf(element.getAction())));
		} else {
			component.setToolTipText(null);
		}

		// Set foreground and background colors:
		boolean warning = element.getAction() == SyncAction.ASK_LENGTH_INCONSISTENT
				|| element.getAction() == SyncAction.ASK_FILES_GT_HISTORY
				|| element.getAction() == SyncAction.ASK_FILES_NOT_IN_HISTORY;
		if (isSelected) {
			if (warning) {
				component.setForeground(WARNING_SELECTED);
			} else {
				component.setForeground(table.getSelectionForeground());
			}
			component.setBackground(table.getSelectionBackground());
		} else {
			if (warning) {
				component.setForeground(WARNING);
			} else {
				component.setForeground(table.getForeground());
			}
			component.setBackground(table.getBackground());

			// Color the comparison row according to the content:
			if (column == 3) {
				component.setBackground(actionBackgrounds.get(element
						.getAction()));
			}

			// If row contains a directory use different color:
			if (element.isDirectory()
					&& component.getBackground().equals(table.getBackground()))
				component.setBackground(DIRECTORY);
		}

		if (!element.isActive()) {
			component.setEnabled(false);
		} else {
			component.setEnabled(true);
		}

		return component;
	}

	/**
	 * @see MouseListener#mouseClicked(MouseEvent)
	 */
	public void mouseClicked(MouseEvent e) {
		handlePopupMenu(e);
	}

	/**
	 * @see MouseListener#mouseEntered(MouseEvent)
	 */
	public void mouseEntered(MouseEvent e) {
	}

	/**
	 * @see MouseListener#mouseExited(MouseEvent)
	 */
	public void mouseExited(MouseEvent e) {
	}

	/**
	 * @see MouseListener#mousePressed(MouseEvent)
	 */
	public void mousePressed(MouseEvent e) {
		handlePopupMenu(e);
	}

	/**
	 * @see MouseListener#mouseReleased(MouseEvent)
	 */
	public void mouseReleased(MouseEvent e) {
		handlePopupMenu(e);
	}

	/**
	 * Creates a popup menu for a certain mouse event on the table.
	 * 
	 * @param e
	 *            The mouse event to deal with.
	 */
	private void handlePopupMenu(MouseEvent e) {
		if (e.isPopupTrigger()
				&& !localTable.getSelectionModel().isSelectionEmpty()) {
			// Create menu:
			JPopupMenu popup = new JPopupMenu();
			popup.add(JFSSupport.getMenuItem("COMP_TABLE_ACTIVATE_ACTION",
					"ACTIVATE", this));
			popup.add(JFSSupport.getMenuItem("COMP_TABLE_DEACTIVATE_ACTION",
					"DEACTIVATE", this));
			popup.addSeparator();

			// Updates the selection of JFS elements:
			updateSelection();

			// Go through selection and add actions that are common for all
			// selected JFS elements:
			TreeMap<SyncAction, Integer> validActions = new TreeMap<SyncAction, Integer>();
			for (JFSElement element : currentSelection) {
				for (SyncAction a : element.getValidActions()) {
					Integer i = validActions.get(a);
					if (i != null) {
						validActions.put(a, i + 1);
					} else {
						validActions.put(a, 1);
					}
				}
			}

			// If the number of appearances in the hash map is equal to
			// the size of the selection, the action is valid:
			for (Entry<SyncAction, Integer> entry : validActions.entrySet()) {
				if (entry.getValue() == currentSelection.size()) {
					popup.add(JFSSupport.getMenuItem(String.valueOf(entry
							.getKey()), String.valueOf(entry.getKey()), this,
							actionIconNames.get(entry.getKey())));
				}
			}

			// Show menu:
			popup.show(e.getComponent(), e.getX(), e.getY());
		}
	}

	/**
	 * @see ActionListener#actionPerformed(ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		if (localTable.getSelectionModel().isSelectionEmpty())
			return;

		String cmd = e.getActionCommand();

		for (JFSElement element : currentSelection) {
			// Set whether the action of the element is active:
			if (cmd.equals("ACTIVATE")) {
				element.setActive(true);
			}
			if (cmd.equals("DEACTIVATE")) {
				element.setActive(false);
			}

			// Change action based on command string:
			for (SyncAction a : SyncAction.values()) {
				if (cmd.equals(String.valueOf(a)))
					element.setAction(a);
			}
		}

		localTable.updateUI();
	}

	/**
	 * Updates the vector of selected JFS elements based on the selection model.
	 */
	private void updateSelection() {
		ListSelectionModel sm = localTable.getSelectionModel();
		currentSelection.clear();
		for (int i = sm.getMinSelectionIndex(); i <= sm.getMaxSelectionIndex(); i++) {
			if (sm.isSelectedIndex(i)) {
				currentSelection.add(task.getViewTable().get(i));
			}
		}
	}
}