/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.border.TitledBorder;

import jfs.conf.JFSConst;
import jfs.conf.JFSText;
import jfs.sync.JFSProgress;
import jfs.sync.JFSProgressObserver;
import jfs.sync.JFSTask;
import jfs.sync.JFSProgress.ProgressState;

/**
 * A dialog that is shown during the comparison of a set of directory pairs und
 * during the synchronization.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSProgressView.java,v 1.20 2005/10/07 17:43:21 heidrich Exp $
 */
public class JFSProgressView implements JFSProgressObserver, ActionListener {

	/** The main frame ;-). */
	private JFSMainView mainView;

	/** The progress dialog. */
	private JDialog dialog;

	/** The task object. */
	private final JFSTask task;

	/** The completion bar for the algorithm. */
	private JProgressBar completionBar;

	/** The file transfer bar for the algorithm. */
	private JProgressBar fileTransferBar;

	/** The area with the status information. */
	private JTextArea commands;

	/** The detailed view on the performed commands. */
	private JPanel commandsPanel;

	/** Shows and hides synchronization details. */
	private JButton toggleDetailsButton;

	/**
	 * Initializes the progress view.
	 * 
	 * @param mainView
	 *            The main frame.
	 * @param task
	 *            The task object.
	 */
	public JFSProgressView(JFSMainView mainView, JFSTask task) {
		this.mainView = mainView;
		this.task = task;

		// Get the translation object:
		JFSText t = JFSText.getInstance();

		// Create the modal dialog:
		dialog = new JDialog(mainView.getFrame(), true);
		dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		dialog.setResizable(false);

		Container cp = dialog.getContentPane();
		cp.setLayout(new BorderLayout());

		// Create state panel:
		JPanel statePanel = new JPanel(new BorderLayout());
		statePanel.setBorder(new TitledBorder(t.get("PROGRESS_STATE")));
		completionBar = new JProgressBar(JProgressBar.HORIZONTAL, 0, 100);
		completionBar.setStringPainted(false);
		statePanel.add(completionBar, BorderLayout.NORTH);
		JPanel fileTransferPanel = new JPanel();
		fileTransferBar = new JProgressBar(JProgressBar.HORIZONTAL, 0, 100);
		fileTransferBar.setStringPainted(false);
		fileTransferBar.setVisible(false);
		fileTransferPanel.add(fileTransferBar);
		statePanel.add(fileTransferPanel, BorderLayout.WEST);

		JPanel detailsPanel = new JPanel();
		toggleDetailsButton = JFSSupport.getButton("PROGRESS_SHOW_DETAILS",
				"TOGGLE_DETAILS", this);
		detailsPanel.add(toggleDetailsButton);
		statePanel.add(toggleDetailsButton, BorderLayout.EAST);
		Dimension d = statePanel.getPreferredSize();
		statePanel.setPreferredSize(new Dimension(280, d.height));

		// Create command panel:
		commandsPanel = new JPanel(new GridLayout(1, 1));
		commandsPanel.setBorder(new TitledBorder(t
				.get("PROGRESS_DETAILS_TITLE")));
		commands = new JTextArea();
		commands.setEnabled(false);
		commandsPanel.add(new JScrollPane(commands));
		commandsPanel.setPreferredSize(new Dimension(280, 150));

		// Create cancel button in a separate panel:
		JPanel buttonPanel = new JPanel();
		buttonPanel.add(JFSSupport.getButton("CANCEL", "CANCEL", this));

		// Add all panels:
		cp.add(statePanel, BorderLayout.NORTH);
		cp.add(commandsPanel, BorderLayout.CENTER);
		cp.add(buttonPanel, BorderLayout.SOUTH);

		// Attach as observer:
		task.getProgress().attach(this);
	}

	/**
	 * Starts a new thread for the computation of comparison, synchronization,
	 * or synchronization with testrun of two file hierarchies.
	 * 
	 * @param mode
	 *            The chosen mode.
	 */
	public void compute(final ProgressState mode) {
		// Center the dialog before each computation to reflect changes of
		// the main frame:
		commandsPanel.setVisible(false);
		fileTransferBar.setVisible(false);
		dialog.pack();
		JFSSupport.center(mainView.getFrame(), dialog);
		completionBar.setValue(0);
		fileTransferBar.setValue(0);

		// Get the translation object:
		JFSText t = JFSText.getInstance();

		// Create the title:
		String title = "";

		if (mode == ProgressState.COMPARISON) {
			title = t.get("COMPARISON");
		} else if (mode == ProgressState.SYNCHRONIZATION) {
			title = t.get("SYNCHRONIZATION");
		}

		dialog.setTitle(title);

		// Set initial label text:
		commands.setText("");
		commands.append(t.get("PROGRESS_INIT"));
		commands.append(System.getProperty("line.separator"));

		// Define new thread for the computation:
		final Thread thread = new Thread() {

			public void run() {
				// Wait for dialog to appear:
				while (!dialog.isVisible()) {
				}

				// Chose right operation:
				if (mode == JFSProgress.ProgressState.COMPARISON) {
					task.compare();
				} else if (mode == JFSProgress.ProgressState.SYNCHRONIZATION) {
					task.synchronize();
				}

				// Update frame's UI:
				mainView.update();

				// Dispose dialog:
				dialog.setVisible(false);
				dialog.dispose();
			}
		};

		// Start thread:
		thread.start();

		// Make dialog window visible:
		dialog.setVisible(true);
	}

	/**
	 * @see JFSProgressObserver#updateFileTransferState(JFSProgress)
	 */
	public void updateFileTransferState(JFSProgress progress) {
		// Update size information:
		long bytesTransfered = progress.getBytesTransfered();
		long bytesToTransfer = progress.getBytesToTransfer();

		if (bytesToTransfer > JFSConst.VIEW_MIN_BYTES && bytesTransfered > 0) {
			fileTransferBar.setValue(progress.getFileTransferRatio());
			fileTransferBar.setVisible(true);
		} else {
			fileTransferBar.setVisible(false);
		}
	}

	/**
	 * @see JFSProgressObserver#updateCompletionState(JFSProgress)
	 */
	public void updateCompletionState(JFSProgress progress) {
		// Get the translation object:
		JFSText t = JFSText.getInstance();

		// Update the progress bar:
		completionBar.setValue(progress.getCompletionRatio());

		// Update the command string currently performed:
		String stateString = t.get("PROGRESS_PROC");
		String connector = "";

		ProgressState state = progress.getState();
		if (state == ProgressState.COMPARISON) {
			stateString = t.get("PROGRESS_COMP");
			connector = t.get("PROGRESS_COMP_CON");
		} else if (state == ProgressState.SYNCHRONIZATION) {
			stateString = t.get("PROGRESS_SYNC");
		} else if (state == ProgressState.SYNCHRONIZATION_COPY) {
			stateString = t.get("PROGRESS_COPY");
			connector = t.get("PROGRESS_COPY_CON");
		} else if (state == ProgressState.SYNCHRONIZATION_DELETE) {
			stateString = t.get("PROGRESS_DELETE");
		}

		String currentSrc = progress.getCurrentSrc();
		String currentTgt = progress.getCurrentTgt();

		if (currentSrc != null)
			stateString += (" <" + currentSrc + ">");

		if ((currentSrc != null) && (currentTgt != null))
			stateString += (" " + connector);

		if (currentTgt != null)
			stateString += (" <" + currentTgt + ">");

		commands.append(stateString);
		commands.append(System.getProperty("line.separator"));

		mainView.updateComparisonTable();
	}

	/**
	 * @see ActionListener#actionPerformed(ActionEvent)
	 */
	public void actionPerformed(ActionEvent event) {
		String cmd = event.getActionCommand();

		if (cmd.equals("TOGGLE_DETAILS")) {
			JFSText t = JFSText.getInstance();

			if (commandsPanel.isVisible()) {
				toggleDetailsButton.setText(t.get("PROGRESS_SHOW_DETAILS"));
				commandsPanel.setVisible(false);
			} else {
				toggleDetailsButton.setText(t.get("PROGRESS_HIDE_DETAILS"));
				commandsPanel.setVisible(true);
			}

			dialog.pack();
			JFSSupport.center(dialog.getParent(), dialog);
		}

		if (cmd.equals("CANCEL")) {
			task.getProgress().cancel();
		}
	}
}