/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.shell;

import java.io.PrintStream;
import java.util.Vector;

import jfs.conf.JFSLog;
import jfs.conf.JFSText;
import jfs.conf.JFSSyncMode.SyncAction;
import jfs.sync.JFSCopyStatement;
import jfs.sync.JFSDeleteStatement;
import jfs.sync.JFSElement;
import jfs.sync.JFSFile;

/**
 * This class performs all necessary formatting operations for printing the
 * comparison table and the copy and delete statements.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSPrint.java,v 1.4 2005/10/04 19:32:38 heidrich Exp $
 */
public class JFSPrint {

	/**
	 * Adapts a string according to the maximum number of displayable
	 * characters. If the original string has to many chars, the trailing chars
	 * are cut, if it has to few chars, spaces are added. Finally the output
	 * string is set in quotations. If the input string is null, a string of
	 * spaces is returned.
	 * 
	 * @param in
	 *            The input string.
	 * @param max
	 *            The maximum length.
	 * @return The modified string of length max.
	 */
	public static String adapt(String in, int max) {
		if (in == null) {
			char[] spaces = new char[max];
			for (int i = 0; i < max; i++)
				spaces[i] = ' ';
			return new String(spaces);
		}

		int length = in.length();
		String out;

		if (length < max - 2) {
			char[] spaces = new char[max - 2 - length];

			for (int i = 0; i < max - 2 - length; i++)
				spaces[i] = ' ';

			out = "'" + in + "'" + new String(spaces);
		} else if (length > max - 2) {
			out = "'" + in.substring(length - max + 2) + "'";
		} else {
			out = "'" + in + "'";
		}

		return out;
	}

	/**
	 * Prints the comparison table.
	 * 
	 * @param table
	 *            The table to be printed.
	 */
	public static void printComparisonTable(Vector<JFSElement> table) {
		// Get translation object:
		JFSText t = JFSText.getInstance();

		PrintStream out = JFSLog.getOut().getStream();
		out.println(t.get("PRINT_COMPARISON_TITLE"));

		for (JFSElement element : table) {
			JFSFile src = element.getSrcFile();
			JFSFile tgt = element.getTgtFile();

			// Compute source string:
			String srcString = null;

			if (src != null)
				if (element.isDirectory())
					srcString = src.getPath();
				else
					srcString = src.getName();

			if (element.isDirectory())
				srcString = JFSPrint.adapt(srcString, 32);
			else
				srcString = "  " + JFSPrint.adapt(srcString, 30);

			// Compute action string:
			String actionString;

			if (element.getAction() == SyncAction.NOP)
				actionString = "   ";
			else if (element.getAction() == SyncAction.COPY_SRC)
				actionString = " > ";
			else if (element.getAction() == SyncAction.COPY_TGT)
				actionString = " < ";
			else if (element.getAction() == SyncAction.DELETE_SRC
					|| element.getAction() == SyncAction.DELETE_TGT
					|| element.getAction() == SyncAction.DELETE_SRC_AND_TGT)
				actionString = " - ";
			else
				actionString = " ? ";

			// Compute target string:
			String tgtString = null;

			if (tgt != null)
				if (element.isDirectory())
					tgtString = tgt.getPath();
				else
					tgtString = tgt.getName();

			if (element.isDirectory())
				tgtString = JFSPrint.adapt(tgtString, 32);
			else
				tgtString = "  " + JFSPrint.adapt(tgtString, 30);

			// Print row:
			out.println(srcString + actionString + tgtString);
		}

		out.println(t.get("PRINT_COMPARISON_HELP"));
		out.println(t.get("PRINT_COUNT") + " " + table.size());
		out.println();
	}

	/**
	 * Prints all copy statements.
	 * 
	 * @param copyStatements
	 *            The copy statements to be printed.
	 */
	public static void printCopyStatements(
			Vector<JFSCopyStatement> copyStatements) {
		// Get translation object:
		JFSText t = JFSText.getInstance();

		PrintStream out = JFSLog.getOut().getStream();
		out.println(t.get("PRINT_COPY_FILE_TITLE"));

		// Print information about the files to copy:
		int i = 0;
		for (JFSCopyStatement cs : copyStatements) {
			String srcString = JFSPrint.adapt(cs.getSrc().getPath(), 27);
			String tgtString = JFSPrint.adapt(cs.getTgt().getPath(), 27);
			String copyFlag;

			if (cs.getCopyFlag())
				copyFlag = "X";
			else
				copyFlag = " ";

			out.println("[" + copyFlag + "] " + t.get("PRINT_COPY_FILE_CMD")
					+ " " + srcString + " "
					+ t.get("PRINT_COPY_FILE_CONNECTOR") + " " + tgtString
					+ " (" + t.get("PRINT_NUMBER") + " " + (i + 1) + ")");
			i++;
		}

		out.println(t.get("PRINT_COUNT") + " " + copyStatements.size());
		out.println();
	}

	/**
	 * Prints all delete statements.
	 * 
	 * @param deleteStatements
	 *            The delete statements to be printed.
	 */
	public static void printDeleteStatements(
			Vector<JFSDeleteStatement> deleteStatements) {
		// Get translation object:
		JFSText t = JFSText.getInstance();

		PrintStream out = JFSLog.getOut().getStream();
		out.println(t.get("PRINT_DELETE_FILE_TITLE"));

		// Print information about the files to delete:
		int i = 0;
		for (JFSDeleteStatement ds : deleteStatements) {
			String fileString = JFSPrint.adapt(ds.getFile().getPath(), 56);
			String deleteFlag;

			if (ds.getDeleteFlag())
				deleteFlag = "X";
			else
				deleteFlag = " ";

			out.println("[" + deleteFlag + "] "
					+ t.get("PRINT_DELETE_FILE_CMD") + " " + fileString + " ("
					+ t.get("PRINT_NUMBER") + " " + (i + 1) + ")");
			i++;
		}

		out.println(t.get("PRINT_COUNT") + " " + deleteStatements.size());
		out.println();
	}

	/**
	 * Prints information about files that have not been copied; i.e., the copy
	 * statement was not performed successfully. (Only files with an active copy
	 * flag are shown.) If the vector is empty nothing is printed.
	 * 
	 * @param copyStatements
	 *            The vector to be printed.
	 */
	public static void printFailedCopyStatements(
			Vector<JFSCopyStatement> copyStatements) {
		if (copyStatements.isEmpty())
			return;

		// Get translation object:
		JFSText t = JFSText.getInstance();

		PrintStream out = JFSLog.getOut().getStream();
		out.println(t.get("PRINT_NOT_COPIED_FILES_TITLE"));

		// Print information about not copied files:
		for (JFSCopyStatement cs : copyStatements) {
			String srcString = JFSPrint.adapt(cs.getSrc().getPath(), 27);
			String tgtString = JFSPrint.adapt(cs.getTgt().getPath(), 27);
			out.println(t.get("PRINT_COPY_FILE_CMD") + " " + srcString + " "
					+ t.get("PRINT_COPY_FILE_CONNECTOR") + " " + tgtString);
		}

		out.println(t.get("PRINT_COUNT") + " " + copyStatements.size());
		out.println();
	}

	/**
	 * Prints information about files that have not been deleted; i.e., the
	 * delete statement was not performed successfully. (Only files with active
	 * delete flag are shown.) If the vector is empty nothing is printed.
	 * 
	 * @param deleteStatements
	 *            The vector to be printed.
	 */
	public static void printFailedDeleteStatements(
			Vector<JFSDeleteStatement> deleteStatements) {
		if (deleteStatements.isEmpty())
			return;

		// Get translation object:
		JFSText t = JFSText.getInstance();

		PrintStream out = JFSLog.getOut().getStream();
		out.println(t.get("PRINT_NOT_DELETED_FILES_TITLE"));

		// Print information about not deleted files:
		for (JFSDeleteStatement ds : deleteStatements) {
			String fileString = JFSPrint.adapt(ds.getFile().getPath(), 56);
			out.println(t.get("PRINT_DELETE_FILE_CMD") + " " + fileString);
		}

		out.println(t.get("PRINT_COUNT") + " " + deleteStatements.size());
		out.println();
	}
}