/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.shell;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;

import jfs.conf.JFSLog;
import jfs.conf.JFSText;
import jfs.sync.JFSProgress;
import jfs.sync.JFSProgressObserver;
import jfs.sync.JFSProgress.ProgressState;

/**
 * A simple observer of the algorithm's progress that prints the state to
 * standard out.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSProgressPrint.java,v 1.4 2005/10/04 19:32:38 heidrich Exp $
 */
public class JFSProgressPrint implements JFSProgressObserver {

	/** The stream for controling cancel operations. */
	private BufferedReader din = new BufferedReader(new InputStreamReader(
			System.in));

	/**
	 * @see JFSProgressObserver#updateFileTransferState(JFSProgress)
	 */
	public void updateFileTransferState(JFSProgress progress) {
		// Do not display anything when the file transfer state is changed.
	}

	/**
	 * @see JFSProgressObserver#updateCompletionState(JFSProgress)
	 */
	public void updateCompletionState(JFSProgress progress) {
		// Get the translation object:
		JFSText t = JFSText.getInstance();

		// Update the command string currently performed:
		String stateString = t.get("PROGRESS_PROC");
		String connector = "";

		ProgressState state = progress.getState();
		if (state == ProgressState.COMPARISON) {
			stateString = t.get("PROGRESS_COMP");
			connector = t.get("PROGRESS_COMP_CON");
		} else if (state == ProgressState.SYNCHRONIZATION) {
			stateString = t.get("PROGRESS_SYNC");
		} else if (state == ProgressState.SYNCHRONIZATION_COPY) {
			stateString = t.get("PROGRESS_COPY");
			connector = t.get("PROGRESS_COPY_CON");
		} else if (state == ProgressState.SYNCHRONIZATION_DELETE) {
			stateString = t.get("PROGRESS_DELETE");
		}

		String currentSrc = progress.getCurrentSrc();
		String currentTgt = progress.getCurrentTgt();

		if (currentSrc != null)
			stateString += (" " + JFSPrint.adapt(currentSrc, 25));

		if ((currentSrc != null) && (currentTgt != null))
			stateString += (" " + connector);

		if (currentTgt != null)
			stateString += (" " + JFSPrint.adapt(currentTgt, 25));

		int percentageValue = progress.getCompletionRatio();
		String percentage = String.valueOf(percentageValue);

		if (percentageValue < 10)
			percentage = "  " + percentage;
		else if (percentageValue < 100)
			percentage = " " + percentage;

		PrintStream out = JFSLog.getOut().getStream();
		out.print("[" + percentage + "%] " + stateString + " ");
		out.println();

		// Ask for canceling the algorithm if 'return' is entered:
		try {
			if (din.ready()) {
				// Skip first enter:
				din.readLine();

				// Read answer:
				String input = "";
				String yes = t.get("PROGRESS_CANCEL_YES");
				String no = t.get("PROGRESS_CANCEL_NO");

				while (!input.equals(yes) && !input.equals(no)) {
					out.println(t.get("PROGRESS_CANCEL"));
					out.print(t.get("PROGRESS_CANCEL_INPUT") + " ");
					input = din.readLine().toLowerCase();
				}

				if (input.equals(yes)) {
					out.println(t.get("PROGRESS_CANCEL_REQUEST"));
					out.println();
					progress.cancel();
				} else
					out.println();
			}
		} catch (IOException e) {
			// Just continue...
		}
	}
}