/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync;

import java.io.PrintStream;
import java.util.Vector;

import jfs.conf.JFSConfig;
import jfs.conf.JFSDirectoryPair;
import jfs.conf.JFSHistory;
import jfs.conf.JFSHistoryManager;
import jfs.conf.JFSLog;
import jfs.conf.JFSText;
import jfs.shell.JFSPrint;
import jfs.shell.JFSProgressPrint;
import jfs.sync.JFSProgress.ProgressState;

/**
 * Computes a comparison table from a JFS directory pair that represents the
 * common and the different parts for the two directories that have to be
 * compared. Furthermore, file factories are created for the source and target
 * side, which may be accessed (and shut down) via this object.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSComparison.java,v 1.25 2005/10/04 19:32:38 heidrich Exp $
 */
class JFSComparison {

	/** Monitors the algorithm's progress. */
	private JFSProgress progress;

	/**
	 * Represents the resulting comparison table of JFSElement objects, that is
	 * all mapped files and directories.
	 */
	private Vector<JFSElement> table = new Vector<JFSElement>();

	/** The used producer to create source JFS file objects. */
	private JFSFileProducer srcProducer;

	/** The used producer to create target JFS file objects. */
	private JFSFileProducer tgtProducer;

	/** The synchronization history. */
	private JFSHistory history;

	/**
	 * Creates a new synchronization object for a certain directory pair.
	 * 
	 * @param pair
	 *            The directory pair to deal with.
	 * @param progress
	 *            The used progress object.
	 */
	public JFSComparison(JFSDirectoryPair pair, JFSProgress progress) {
		JFSFileProducerManager pm = JFSFileProducerManager.getInstance();
		srcProducer = pm.createProducer(pair.getSrc());
		tgtProducer = pm.createProducer(pair.getTgt());
		this.progress = progress;
		history = JFSHistoryManager.getInstance().getHistory(pair);
		history.load();
	}

	/**
	 * Returns the comparison table itself; i.e., the list of all mapped
	 * elements.
	 * 
	 * @return The list of JFS elements.
	 */
	public final Vector<JFSElement> getTable() {
		return table;
	}

	/**
	 * Returns the assigned source JFS file producer.
	 * 
	 * @return The source producer.
	 */
	public JFSFileProducer getSrcProducer() {
		return srcProducer;
	}

	/**
	 * Returns the assigned target JFS file producer.
	 * 
	 * @return The target producer.
	 */
	public JFSFileProducer getTgtProducer() {
		return tgtProducer;
	}

	/**
	 * Returns the history.
	 * 
	 * @return The history.
	 */
	public JFSHistory getHistory() {
		return history;
	}

	/**
	 * Inserts an element to the comparison table and starts the comparison
	 * algorithm recursively, if and only if the matched files are directories.
	 * At least one (source or target) file has to be not equal to null.
	 * 
	 * @param srcFile
	 *            A source file which may be null if no corresponding source
	 *            file exists, but a target file.
	 * @param tgtFile
	 *            A target file which may be null if no corresponding target
	 *            file exists, but a source file.
	 * @param isDirectory
	 *            Determines whether the files are directories.
	 */
	private final void add(JFSFile srcFile, JFSFile tgtFile, boolean isDirectory) {
		assert srcFile != null || tgtFile != null;

		// Determine whether the comparison should be performed:
		JFSConfig config = JFSConfig.getInstance();
		if (!config.getIncludes().isEmpty()) {
			if (srcFile != null && !config.matchesIncludes(srcFile.getName()))
				return;
			if (tgtFile != null && !config.matchesIncludes(tgtFile.getName()))
				return;
		}
		if (!config.getExcludes().isEmpty()) {
			if (srcFile != null && config.matchesExcludes(srcFile.getName()))
				return;
			if (tgtFile != null && config.matchesExcludes(tgtFile.getName()))
				return;
		}

		// Add an element to the comparison table:
		table.add(new JFSElement(srcFile, tgtFile, isDirectory));

		// Start algorithm recursivley, if the files are directories:
		// (This is the case, if one of them is a directory!)
		if (isDirectory && !progress.isCanceled())
			compare(srcFile, tgtFile);
	}

	/**
	 * Compares two lists of files and writes the result to the comparison
	 * table. Both lists must be not equal to null.
	 * 
	 * @param srcFiles
	 *            The array of source files.
	 * @param tgtFiles
	 *            The array of target files.
	 * @param isDirectory
	 *            Determines whether the files are directories.
	 */
	private final void compareLists(JFSFile[] srcFiles, JFSFile[] tgtFiles,
			boolean isDirectory) {
		assert srcFiles != null && tgtFiles != null;

		// First, sort the input arrays:
		HeapSort<JFSFile> h = new HeapSort<JFSFile>();
		h.sort(srcFiles);
		h.sort(tgtFiles);

		int srcIndex = 0;
		int tgtIndex = 0;

		while (srcIndex < srcFiles.length && tgtIndex < tgtFiles.length) {
			int comp = srcFiles[srcIndex].compareTo(tgtFiles[tgtIndex]);

			if (comp == 0) {
				// Case 1: We found two matching files:
				add(srcFiles[srcIndex], tgtFiles[tgtIndex], isDirectory);
				srcIndex++;
				tgtIndex++;
			} else if (comp > 0) {
				// Case 2: No matching file was found and the source file is
				// greater than the target file. In this case we have to write
				// the target file to the comparison table and investigate the
				// next target file in the list:
				add(null, tgtFiles[tgtIndex], isDirectory);
				tgtIndex++;
			} else if (comp < 0) {
				// Case 3: No matching file was found and the target file is
				// greater than the source file. In this case we have to write
				// the source file to the comparison table and investigate the
				// next source file in the list:
				add(srcFiles[srcIndex], null, isDirectory);
				srcIndex++;
			}
		}

		// Case 4: All source files were already handled. In this case we have
		// to write the rest of the target files into the table:
		while (tgtIndex < tgtFiles.length) {
			add(null, tgtFiles[tgtIndex], isDirectory);
			tgtIndex++;
		}

		// Case 5: All target files were already handled. In this case we have
		// to write the rest of the source files into the table:
		while (srcIndex < srcFiles.length) {
			add(srcFiles[srcIndex], null, isDirectory);
			srcIndex++;
		}
	}

	/**
	 * Compares two directories and writes the result to the comparison table.
	 * At least one (source or target) file has to be not equal to null.
	 * 
	 * @param srcDir
	 *            A source directory which may be null if no corresponding
	 *            source directory exists, but a target directory.
	 * @param tgtDir
	 *            A target directory which may be null if no corresponding
	 *            target directory exists, but a source directory.
	 */
	private final void compare(JFSFile srcDir, JFSFile tgtDir) {
		assert srcDir != null || tgtDir != null;

		JFSFile[] srcFileList = null;
		JFSFile[] tgtFileList = null;
		JFSFile[] srcDirectoryList = null;
		JFSFile[] tgtDirectoryList = null;
		String currentSrc = null;
		String currentTgt = null;
		int weight = 0;

		if (srcDir != null) {
			srcFileList = srcDir.getFileList();
			srcDirectoryList = srcDir.getDirectoryList();
			currentSrc = srcDir.getPath();
			weight += 1;
		}

		if (tgtDir != null) {
			tgtFileList = tgtDir.getFileList();
			tgtDirectoryList = tgtDir.getDirectoryList();
			currentTgt = tgtDir.getPath();
			weight += 1;
		}

		// Create dummy lists if null was returned:
		if (srcFileList == null)
			srcFileList = new JFSFile[0];

		if (tgtFileList == null)
			tgtFileList = new JFSFile[0];

		if (srcDirectoryList == null)
			srcDirectoryList = new JFSFile[0];

		if (tgtDirectoryList == null)
			tgtDirectoryList = new JFSFile[0];

		// Update the progress object:
		progress.setCurrentSrc(currentSrc);
		progress.setCurrentTgt(currentTgt);
		progress.increase(srcDirectoryList.length + tgtDirectoryList.length,
				weight);
		progress.fireCompletionStateUpdate();

		compareLists(srcFileList, tgtFileList, false);
		compareLists(srcDirectoryList, tgtDirectoryList, true);

		progress.decrease();
	}

	/**
	 * Compares the two directories of the JFS directory pair and creates the
	 * comparison table.
	 */
	public final void compare() {
		// Get translation object:
		JFSText t = JFSText.getInstance();

		// Clear vector first:
		table.clear();

		// Update the progress object:
		progress.setState(ProgressState.COMPARISON);
		progress.increase(2, 1);

		long time = System.currentTimeMillis();

		// Create JFS files for the source and target directories to handle:
		JFSFile srcFile = srcProducer.getRootJfsFile();
		JFSFile tgtFile = tgtProducer.getRootJfsFile();

		// Test whether the file objects exists and are directories:
		if (!srcFile.exists() || !tgtFile.exists() || !srcFile.isDirectory()
				|| !tgtFile.isDirectory()) {
			PrintStream p = JFSLog.getErr().getStream();
			p.println(t.get("ERROR_VALID_DIRECTORY_PAIR"));
			p.println("  '" + srcProducer.getRootPath() + "', ");
			p.println("  '" + tgtProducer.getRootPath() + "'");
		} else {
			// Add root element to the comparison table:
			table.add(new JFSElement(srcFile, tgtFile, true));
			// Start the comparison:
			compare(srcFile, tgtFile);
		}

		// Update the progress object:
		progress.decrease();
		progress.setDuration(System.currentTimeMillis() - time);
		progress.fireCompletionStateUpdate();
		progress.finish();
	}

	/**
	 * Shut downs the file producers of the comparison object when not needed
	 * any more (e.g., before creating a new one for the directory pair).
	 */
	public final void shutDownProducers() {
		JFSFileProducerManager pm = JFSFileProducerManager.getInstance();
		pm.shutDownProducer(srcProducer.getUri());
		pm.shutDownProducer(tgtProducer.getUri());
	}

	/**
	 * Tests the algorithm.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		String srcDir = "test/dirs/source";
		String tgtDir = "test/dirs/target";

		String usage = "Usage:\n" + "  JFSComparison [options]\n"
				+ "Options:\n" + "  -src <directory>  The source directory.\n"
				+ "  -tgt <directory>  The target directory.";

		int i = 0;

		try {
			while (i < args.length) {
				if (args[i].equals("-src")) {
					i++;
					srcDir = args[i];
				} else if (args[i].equals("-tgt")) {
					i++;
					tgtDir = args[i];
				} else {
					JFSLog.getOut().getStream().println(usage);
					System.exit(0);
				}

				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			JFSLog.getErr().getStream().println(usage);
			System.exit(0);
		}

		JFSProgress progress = new JFSProgress();
		JFSProgressPrint observer = new JFSProgressPrint();
		JFSDirectoryPair pair = new JFSDirectoryPair(srcDir, tgtDir);
		JFSComparison comparison = new JFSComparison(pair, progress);

		progress.attach(observer);
		comparison.compare();
		JFSPrint.printComparisonTable(comparison.getTable());
	}
}