/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync;

import java.util.TreeSet;

import jfs.conf.JFSConfig;
import jfs.conf.JFSSyncModes;
import jfs.conf.JFSSyncMode.SyncAction;

/**
 * Represents a single pair of corresponding files or directories on the source
 * and target side within the directory structures that have to be compared.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSElement.java,v 1.17 2005/10/04 19:32:38 heidrich Exp $
 */
public class JFSElement implements Comparable<JFSElement> {

	/** The states of the element. */
	public enum ElementState {
		NOT_DETERMINED, EQUAL, SRC_IS_NULL, TGT_IS_NULL, SRC_GT_TGT, TGT_GT_SRC, LENGTH_INCONSISTENT
	};

	/** Source file. */
	private JFSFile srcFile;

	/** Target file. */
	private JFSFile tgtFile;

	/** Determines whether the File objects are directories. */
	private boolean isDirectory;

	/** Determines the state of the element. */
	private ElementState state = ElementState.NOT_DETERMINED;

	/**
	 * The action that has to be performed for the JFS element.
	 * 
	 * @see JFSSyncModes
	 */
	private SyncAction action = SyncAction.NOP;

	/** Determines whether the action is active or should be skipped. */
	private boolean isActive = true;

	/**
	 * Default constructor: Compares two File objects and determines whether one
	 * is newer than the other or whether both files are equal. At least one
	 * file has to be not equal to null. If both files are equal, there relative
	 * paths have to match.
	 * 
	 * @param srcFile
	 *            Source file.
	 * @param tgtFile
	 *            Target file.
	 * @param isDirectory
	 *            Determines whether the File objects are directories.
	 */
	public JFSElement(JFSFile srcFile, JFSFile tgtFile, boolean isDirectory) {
		this.srcFile = srcFile;
		this.tgtFile = tgtFile;
		this.isDirectory = isDirectory;

		assert srcFile != null || tgtFile != null;

		revalidate();
	}

	/**
	 * Revalidates the comparison table element; that is checks which file is
	 * newer, etc.
	 */
	public final void revalidate() {
		// Reset attributes:
		state = ElementState.NOT_DETERMINED;

		// Get last modified information:
		long srcLastModified = 0;
		long tgtLastModified = 0;

		if (!isDirectory) {
			if (srcFile != null)
				srcLastModified = srcFile.getLastModified();
			if (tgtFile != null)
				tgtLastModified = tgtFile.getLastModified();
		}

		// Comparison:
		// Under the DOS and Windows FAT file system, the finest granularity on
		// time resolution is two seconds. So we have a maximum tolerance range
		// of 2000ms for each comparison:
		long diffTime = compareToTime(srcLastModified, tgtLastModified);

		if (srcFile == null)
			state = ElementState.SRC_IS_NULL;
		else if (tgtFile == null)
			state = ElementState.TGT_IS_NULL;
		else if (!isDirectory)
			if (diffTime == 0)
				state = ElementState.EQUAL;
			else if (diffTime > 0)
				state = ElementState.SRC_GT_TGT;
			else
				state = ElementState.TGT_GT_SRC;
		else
			state = ElementState.EQUAL;

		// Check length:
		if (state == ElementState.EQUAL
				&& srcFile.getLength() != tgtFile.getLength())
			state = ElementState.LENGTH_INCONSISTENT;

		// Set action to NOP if isEqual is true:
		if (state == ElementState.EQUAL)
			action = SyncAction.NOP;
	}

	/**
	 * Returns the source file.
	 * 
	 * @return Source file.
	 */
	public final JFSFile getSrcFile() {
		return srcFile;
	}

	/**
	 * Sets the source file.
	 * 
	 * @param file
	 *            The file to set as source.
	 */
	public void setSrcFile(JFSFile file) {
		srcFile = file;
	}

	/**
	 * Returns the target file.
	 * 
	 * @return Target file.
	 */
	public final JFSFile getTgtFile() {
		return tgtFile;
	}

	/**
	 * Sets the traget file.
	 * 
	 * @param file
	 *            The file to set as target.
	 */
	public void setTgtFile(JFSFile file) {
		tgtFile = file;
	}

	/**
	 * Computes whether source and target files are directories.
	 * 
	 * @return True, if the File objects are directories.
	 */
	public final boolean isDirectory() {
		return isDirectory;
	}

	/**
	 * @return Returns the state of the element.
	 */
	public final ElementState getState() {
		return state;
	}

	/**
	 * Returns the action that has to be performed for the JFS element.
	 * 
	 * @see JFSSyncModes
	 * @return The action.
	 */
	public SyncAction getAction() {
		return action;
	}

	/**
	 * Sets the action that has to be performed for the JFS element.
	 * 
	 * @see JFSSyncModes
	 * @param action
	 *            The action to set.
	 */
	public void setAction(SyncAction action) {
		this.action = action;
	}

	/**
	 * Returns whether the action of the JFS element is active.
	 * 
	 * @return True, if the action is active.
	 */
	public boolean isActive() {
		return isActive;
	}

	/**
	 * Sets whether the action of the JFS element is active.
	 * 
	 * @param isActive
	 *            True, if the action is active.
	 */
	public void setActive(boolean isActive) {
		this.isActive = isActive;
	}

	/**
	 * Returns the relative path of the element, which has to be the same for
	 * source and target if both files are equal. If the source is equal to null
	 * the relative path of the target is returned, and vice versa. According to
	 * the constructor of this element, it is not allowed that both files are
	 * equal to null.
	 * 
	 * @return The relative path of the comparison element.
	 */
	public String getRelativePath() {
		if (srcFile != null)
			return srcFile.getRelativePath();
		else
			return tgtFile.getRelativePath();
	}

	/**
	 * @see Comparable#compareTo(Object)
	 */
	public int compareTo(JFSElement e) {
		return getRelativePath().compareTo(e.getRelativePath());
	}

	/**
	 * Compares two time stamps taking into account the specified granularity of
	 * the configuration object.
	 * 
	 * @param time1
	 *            The first time stamp.
	 * @param time2
	 *            The second time stamp.
	 * @return Returns a positive number if the first is newer than the second
	 *         time stamp, 0 if both are equal, and a negative number if the
	 *         second time stamp is newer than the first one.
	 */
	public static long compareToTime(long time1, long time2) {
		return (time1 - time2) / JFSConfig.getInstance().getGranularity();

	}

	/**
	 * Returns the valid actions for the JFS element.
	 * 
	 * @return A vector of valid synchronization actions.
	 */
	public TreeSet<SyncAction> getValidActions() {
		TreeSet<SyncAction> validActions = new TreeSet<SyncAction>();
		validActions.add(SyncAction.NOP);

		if (srcFile != null) {
			validActions.add(SyncAction.COPY_SRC);
		} else {
			validActions.add(SyncAction.DELETE_TGT);
		}

		if (tgtFile != null) {
			validActions.add(SyncAction.COPY_TGT);
		} else {
			validActions.add(SyncAction.DELETE_SRC);
		}

		if (srcFile != null && tgtFile != null) {
			validActions.add(SyncAction.DELETE_SRC_AND_TGT);
		}

		return validActions;
	}
}