/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync;

import java.util.Vector;

import jfs.conf.JFSConst;
import jfs.shell.JFSProgressPrint;

/**
 * Visits selected steps of the file comparison and synchronization algorithm in
 * order to compute the state of the progression.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSProgress.java,v 1.17 2005/10/04 19:32:38 heidrich Exp $
 */
public class JFSProgress {

	/** The states of the progress object. */
	public enum ProgressState {
		INITIALIZATION, COMPARISON, SYNCHRONIZATION, SYNCHRONIZATION_COPY, SYNCHRONIZATION_DELETE
	};

	/** Duration to compute the table. */
	private long duration = 0;

	/**
	 * Tells about the current (source) work item that is handled by the
	 * algorithm.
	 */
	private String currentSrc = null;

	/**
	 * Tells about the current (target) work item that is handled by the
	 * algorithm.
	 */
	private String currentTgt = null;

	/**
	 * Tells about the current state of the algorithm in form of an unique
	 * integer value.
	 */
	private ProgressState state = ProgressState.INITIALIZATION;

	/**
	 * The number of all items that have to be handled in order to build up the
	 * JFSFile hierarchie on every level of the hierarchie.
	 */
	private Vector<Integer> itemsStarted = new Vector<Integer>();

	/**
	 * The number of all currently handled items on every level of the
	 * hierarchie.
	 */
	private Vector<Integer> itemsHandled = new Vector<Integer>();

	/**
	 * The weigth of all currently handled items on every level of the
	 * hierarchie. The default value is '1'.
	 */
	private Vector<Integer> itemsWeight = new Vector<Integer>();

	/** Completion ratio in percentage. */
	private int completionRatio = 0;

	/** Determines whether the algorithm is canceled or not. */
	private boolean isCanceled = false;

	/** Vector with all oberservers of the alogorithm's progress. */
	private Vector<JFSProgressObserver> observers = new Vector<JFSProgressObserver>();

	/** The time when the observers were updated last. */
	private long updateTime = 0;

	/** The number of bytes to transfer for a file. */
	private long bytesToTransfer = 0;

	/** The number of bytes transfered for a file. */
	private long bytesTransfered = 0;

	/**
	 * Returns duration to compute table.
	 * 
	 * @return Duration.
	 */
	public final long getDuration() {
		return duration;
	}

	/**
	 * Sets the duration needed to perform the algoritm.
	 * 
	 * @param duration
	 *            The duration in ms.
	 */
	final void setDuration(long duration) {
		this.duration += duration;
	}

	/**
	 * Returns the number of bytes to transfer for a file.
	 * 
	 * @return The file size in byte.
	 */
	public long getBytesToTransfer() {
		return bytesToTransfer;
	}

	/**
	 * Sets the number of bytes to transfer for a file.
	 * 
	 * @param bytesToTransfer
	 *            The file size in byte.
	 */
	public void setBytesToTransfer(long bytesToTransfer) {
		this.bytesToTransfer = bytesToTransfer;
	}

	/**
	 * Returns the number of bytes transfered for a file.
	 * 
	 * @return The file size in byte.
	 */
	public long getBytesTransfered() {
		return bytesTransfered;
	}

	/**
	 * Sets the number of bytes transfered for a file.
	 * 
	 * @param bytesTransfered
	 *            The file size in byte.
	 */
	public void setBytesTransfered(long bytesTransfered) {
		this.bytesTransfered = bytesTransfered;
	}

	/**
	 * Returns the ratio of the file size already transfered in percent.
	 * 
	 * @return Percentage value.
	 */
	public final int getFileTransferRatio() {
		if (bytesToTransfer > 0)
			return Math.round((float) bytesTransfered / (float) bytesToTransfer
					* 100);
		else
			return 0;
	}

	/**
	 * Returns the ratio of the items already handled in percent.
	 * 
	 * @return Percentage value.
	 */
	public final int getCompletionRatio() {
		// Compute new percentage:
		float newPercentage = 0;
		for (int i = itemsStarted.size() - 1; i >= 0; i--) {
			int started = itemsStarted.get(i);
			int handled = itemsHandled.get(i);
			int weight = itemsWeight.get(i);

			if (started != 0)
				newPercentage = (weight * (handled + newPercentage)) / started;
		}

		// Ensure that the new percentage value is not lesser than
		// the old one:
		int roundedPercentage = Math.round(newPercentage * 100);

		if (roundedPercentage > this.completionRatio)
			completionRatio = roundedPercentage;

		return completionRatio;
	}

	/**
	 * Sets the current (source) work item, usually a directory or file which
	 * currently handled by the algorithm.
	 * 
	 * @param currentSrc
	 *            String of current work item.
	 */
	final void setCurrentSrc(String currentSrc) {
		this.currentSrc = currentSrc;
	}

	/**
	 * Gets the current (source) work item.
	 * 
	 * @return String of current work item.
	 */
	public final String getCurrentSrc() {
		return currentSrc;
	}

	/**
	 * Sets the current (target) work item, usually a directory or file which
	 * currently handled by the algorithm.
	 * 
	 * @param currentTgt
	 *            String of current work item.
	 */
	final void setCurrentTgt(String currentTgt) {
		this.currentTgt = currentTgt;
	}

	/**
	 * Gets the current (target) work item.
	 * 
	 * @return String of current work item.
	 */
	public final String getCurrentTgt() {
		return currentTgt;
	}

	/**
	 * Sets the state.
	 * 
	 * @param state
	 *            Current state.
	 */
	final void setState(ProgressState state) {
		this.state = state;
	}

	/**
	 * Gets the state.
	 * 
	 * @return Ccurrent state.
	 */
	public final ProgressState getState() {
		return state;
	}

	/**
	 * Increases the progress depth by adding new elements for (1) number of
	 * started items, (2) number of handled items, and finally (3) the weight
	 * for the currently handeled items. Always used in combination with
	 * descrease().
	 * 
	 * @param containedItems
	 *            The number of contained sub items in the currently handled
	 *            item.
	 * @param weight
	 *            The weigth of the currently handled item; i.e., the delta by
	 *            which the number of handled items is increased if this item is
	 *            completely handled.
	 */
	final void increase(int containedItems, int weight) {
		itemsStarted.add(containedItems);
		itemsHandled.add(0);
		itemsWeight.add(weight);
	}

	/**
	 * Decreases the depth for the started and handled items. Always used in
	 * combination with increase().
	 */
	final void decrease() {
		// All vector (should ;-) have the same size, so we just
		// use the size of itemsStarted to remove the last elements:
		int size = itemsStarted.size();

		if (size > 1) {
			// Increase the number of handled directories by the weight of
			// the currently finished directory:
			int handled = itemsHandled.get(size - 2);
			int weight = itemsWeight.get(size - 1);
			itemsHandled.set(size - 2, handled + weight);

			// Because the directory (pair) is handled, we can remove
			// the number of started and handled sub directories and the
			// weight:
			itemsStarted.remove(size - 1);
			itemsHandled.remove(size - 1);
			itemsWeight.remove(size - 1);
		} else {
			itemsStarted.clear();
			itemsHandled.clear();
			itemsWeight.clear();
		}
	}

	/**
	 * Starts a new monitored process.
	 */
	final void start() {
		duration = 0;
		currentSrc = null;
		currentTgt = null;
		state = ProgressState.INITIALIZATION;
		itemsStarted.clear();
		itemsHandled.clear();
		itemsWeight.clear();
		completionRatio = 0;
		isCanceled = false;
		updateTime = 0;
		bytesToTransfer = 0;
		bytesTransfered = 0;
	}

	/**
	 * Shows the end of a part of the monitored algorithm.
	 */
	final void finish() {
		currentSrc = null;
		currentTgt = null;
		state = ProgressState.INITIALIZATION;
		bytesToTransfer = 0;
		bytesTransfered = 0;
	}

	/**
	 * Shows the end of the whole monitored algorithm.
	 */
	final void end() {
		finish();
		updateCompletionState();
	}

	/**
	 * This method is called if an object wants to cancel the algorithm. After
	 * this request the algorithm stops as soon as possible at predefined
	 * milestones.
	 */
	public void cancel() {
		isCanceled = true;
	}

	/**
	 * Determines whether the algorithm is canceled or not. Is so, the algorithm
	 * stops as soon as possible at the next predefined milestone.
	 * 
	 * @return True if and only if the algorithm was cancelled.
	 */
	public boolean isCanceled() {
		return isCanceled;
	}

	/**
	 * Attaches an additional observer.
	 * 
	 * @param observer
	 *            The new observer.
	 */
	public final void attach(JFSProgressObserver observer) {
		observers.add(observer);
	}

	/**
	 * Detaches an existing observer.
	 * 
	 * @param observer
	 *            An old observer.
	 */
	public final void detach(JFSProgressObserver observer) {
		observers.remove(observer);
	}

	/**
	 * Updates the file transfer state for all existing observers.
	 */
	final void fireFileTransferUpdate() {
		if ((System.currentTimeMillis() - updateTime) >= JFSConst.PROGRESS_UPDATE) {
			updateFileTransferState();
			updateTime = System.currentTimeMillis();
		}
	}

	/**
	 * Updates the current state of the algorithm for all existing observers.
	 */
	final void fireCompletionStateUpdate() {
		if ((System.currentTimeMillis() - updateTime) >= JFSConst.PROGRESS_UPDATE) {
			updateCompletionState();
			updateTime = System.currentTimeMillis();
		}
	}

	/**
	 * Updates the file transfer state for all existing observers.
	 */
	private final void updateFileTransferState() {
		for (JFSProgressObserver po : observers) {
			po.updateFileTransferState(this);
		}
	}

	/**
	 * Updates the current state of the algorithm for all existing observers.
	 */
	private final void updateCompletionState() {
		for (JFSProgressObserver po : observers) {
			po.updateCompletionState(this);
		}
	}

	/**
	 * Tests the class.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		JFSProgress progress = new JFSProgress();
		JFSProgressPrint po = new JFSProgressPrint();

		progress.attach(po);
		progress.setState(ProgressState.COMPARISON);
		progress.start();
		progress.increase(10, 1);

		long time = 0;

		for (int i = 0; i < 5; i++) {
			progress.setCurrentSrc("Source Item No. " + i);
			progress.setCurrentTgt("Target Item No. " + i);
			progress.increase(10, 2);
			time = System.currentTimeMillis();

			while ((System.currentTimeMillis() - time) < JFSConst.PROGRESS_UPDATE) {
			}

			progress.fireCompletionStateUpdate();

			for (int j = 0; j < 5; j++) {
				progress.setCurrentSrc("Source Sub Item No. " + j);
				progress.setCurrentTgt("Target Sub Item No. " + j);
				progress.increase(0, 2);
				time = System.currentTimeMillis();

				while ((System.currentTimeMillis() - time) < JFSConst.PROGRESS_UPDATE) {
				}

				progress.fireCompletionStateUpdate();
				progress.decrease();
			}

			progress.decrease();
		}

		progress.end();
		progress.decrease();
	}
}