/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync;

import java.util.Vector;

import jfs.conf.JFSConfig;
import jfs.conf.JFSDirectoryPair;
import jfs.conf.JFSHistory;
import jfs.conf.JFSLog;
import jfs.conf.JFSSyncMode;
import jfs.conf.JFSSyncModes;
import jfs.conf.JFSSyncMode.SyncAction;
import jfs.shell.JFSPrint;
import jfs.shell.JFSProgressPrint;
import jfs.sync.JFSProgress.ProgressState;
import jfs.sync.JFSQuestion.QuestionAnswer;

/**
 * This class is responsible for synchronizing a single JFS directory pair. In
 * order to synchronize a directory pair, (1) the comparison of the two
 * directory structures has to be performed resulting in a comparison table, (2)
 * the actions that have to be applied to the comparison table have to be
 * computed according to the chosen synchronization mode, (3) the list of copy
 * and delete statements has to be determined based on the computed actions, and
 * last (4) the synchronization has to be performed according to the list of
 * copy and delete statements.
 * 
 * @see JFSComparison#compare()
 * @see #computeActions()
 * @see #computeSynchronizationLists()
 * @see #synchronize()
 * @author Jens Heidrich
 * @version $Id: JFSSynchronization.java,v 1.27 2005/10/05 22:06:45 heidrich Exp $
 */
class JFSSynchronization {

	/** The used comparison table. */
	private JFSComparison comparison;

	/** Monitors the algorithm's progress. */
	private JFSProgress progress;

	/** Answers questions during before the algorithm is performed. */
	private JFSQuestion question;

	/** The list of copy statements. */
	private Vector<JFSCopyStatement> copyStatements = new Vector<JFSCopyStatement>();

	/** The list of delete statements. */
	private Vector<JFSDeleteStatement> deleteStatements = new Vector<JFSDeleteStatement>();

	/**
	 * Creates a new synchronization object for a certain directory pair.
	 * 
	 * @param pair
	 *            The directory pair to deal with.
	 * @param progress
	 *            The used progress object.
	 */
	public JFSSynchronization(JFSDirectoryPair pair, JFSProgress progress,
			JFSQuestion question) {
		this.comparison = new JFSComparison(pair, progress);
		this.progress = progress;
		this.question = question;
	}

	/**
	 * Returns the JFS comparison object.
	 * 
	 * @return The used JFS comparison object.
	 */
	public final JFSComparison getComparison() {
		return comparison;
	}

	/**
	 * Returns a list of files that have to be copied.
	 * 
	 * @return Vector of JFS copy statements.
	 */
	public final Vector<JFSCopyStatement> getCopyStatements() {
		return copyStatements;
	}

	/**
	 * Returns a list of files that have to be deleted.
	 * 
	 * @return Vector of JFS delete statements.
	 */
	public final Vector<JFSDeleteStatement> getDeleteStatements() {
		return deleteStatements;
	}

	/**
	 * Computes the actions that have to performed for an entry of the
	 * comparison table according to the chosen synchronization mode.
	 * 
	 * @see JFSSyncMode#computeActions(Vector, JFSHistory)
	 */
	public final void computeActions() {
		JFSSyncMode mode = JFSSyncModes.getInstance().get(
				JFSConfig.getInstance().getSyncMode());

		if (mode == null)
			return;

		mode.computeActions(comparison.getTable(), comparison.getHistory());
	}

	/**
	 * Computes two lists of files according to the actions specified in the
	 * comparison table mode. The first list contains copy statements and
	 * represents all files that have to copied from source to target or the
	 * other way around. The second list contains delete statements and
	 * represents all files that have to be deleted. When the lists are
	 * computed, the user is asked some questions regarding JFS elements for
	 * which no action could be computed automatically.
	 */
	public final void computeSynchronizationLists() {
		JFSFileProducer srcProducer = comparison.getSrcProducer();
		JFSFileProducer tgtProducer = comparison.getTgtProducer();
		copyStatements.clear();
		deleteStatements.clear();
		boolean skipAll = false;

		for (JFSElement element : comparison.getTable()) {
			// Check whether element and corresponding action is active:
			if (!element.isActive())
				continue;

			JFSFile srcFile = element.getSrcFile();
			JFSFile tgtFile = element.getTgtFile();

			// Ask the user first:
			if (!skipAll
					&& (element.getAction() == SyncAction.ASK_LENGTH_INCONSISTENT
							|| element.getAction() == SyncAction.ASK_FILES_GT_HISTORY || element
							.getAction() == SyncAction.ASK_FILES_NOT_IN_HISTORY)) {
				QuestionAnswer a = question.answer(element);
				if (a == QuestionAnswer.SKIP_ALL)
					skipAll = true;
			}

			// Case 1: Copy the file on source side to the target side:
			if (element.getAction() == SyncAction.COPY_SRC) {
				JFSFile newFile = tgtProducer.getJfsFile(srcFile
						.getRelativePath());
				copyStatements.add(new JFSCopyStatement(element, srcFile,
						newFile));
			}

			// Case 2: Copy the file on target side to the source side:
			if (element.getAction() == SyncAction.COPY_TGT) {
				JFSFile newFile = srcProducer.getJfsFile(tgtFile
						.getRelativePath());
				copyStatements.add(new JFSCopyStatement(element, tgtFile,
						newFile));
			}

			// Case 3: Delete the file on target side:
			if (element.getAction() == SyncAction.DELETE_TGT
					|| element.getAction() == SyncAction.DELETE_SRC_AND_TGT) {
				deleteStatements.add(new JFSDeleteStatement(element, tgtFile));
			}

			// Case 4: Delete the file on source side:
			if (element.getAction() == SyncAction.DELETE_SRC
					|| element.getAction() == SyncAction.DELETE_SRC_AND_TGT) {
				deleteStatements.add(new JFSDeleteStatement(element, srcFile));
			}
		}

		// The list of delete statements has to be inverted, because the
		// directories must be empty before they can be deleted:
		JFSTask.invert(deleteStatements);
	}

	/**
	 * Synchronizes the directory structures specified in the directory pair if
	 * and only if the test run flag is not set.
	 */
	public final void synchronize() {
		JFSConfig config = JFSConfig.getInstance();
		long time = System.currentTimeMillis();

		// Set information for monitoring the progress:
		progress.setState(ProgressState.SYNCHRONIZATION);
		progress.increase(deleteStatements.size() + copyStatements.size(), 1);

		int i;
		boolean success;

		// Handle all files to delete first:
		// (This is important for DOS and Windows Operating Systems.
		// Let us assume a file on the source side named "TEST.txt"
		// and on the target side "test.txt" and we want to copy files
		// from source and delete from target.
		// Because JFileSync works with a case-sensitive algorithm,
		// the file "TEST.txt" would be copied first and the file
		// "test.txt" would be deleted afterwards if we would
		// perform the copy statements first. Because DOS and Windows
		// don't distinguish between both file names, there wouldn't exist
		// a file with one of names on the target side after the
		// synchronization.)
		progress.setState(ProgressState.SYNCHRONIZATION_DELETE);
		i = 0;

		while (i < deleteStatements.size() && !progress.isCanceled()) {
			JFSDeleteStatement ds = (JFSDeleteStatement) deleteStatements
					.elementAt(i);
			progress.setCurrentSrc(ds.getFile().getPath());
			progress.setCurrentTgt(null);
			progress.increase(0, 1);
			progress.fireCompletionStateUpdate();

			// Delete only if the delete flag is set, the success flag
			// is false and we are not in testrun mode:
			if (ds.getDeleteFlag() && !ds.getSuccess() && !config.getTestRun()) {
				success = ds.getFile().delete();
				ds.setSuccess(success);

				// Remove from table if action was successfully performed:
				if (success)
					comparison.getTable().remove(ds.getElement());
			}

			progress.decrease();
			i++;
		}

		// Handle all files to copy:
		progress.setState(ProgressState.SYNCHRONIZATION_COPY);
		i = 0;

		while (i < copyStatements.size() && !progress.isCanceled()) {
			JFSCopyStatement cs = (JFSCopyStatement) copyStatements
					.elementAt(i);
			progress.setCurrentSrc(cs.getSrc().getPath());
			progress.setCurrentTgt(cs.getTgt().getPath());
			progress.increase(0, 1);
			progress.fireCompletionStateUpdate();

			// Copy only if the copy flag is set, the success flag
			// is false and we are not in testrun mode:
			if (cs.getCopyFlag() && !cs.getSuccess() && !config.getTestRun()) {
				cs.getSrc().setProgress(progress);
				cs.getTgt().setProgress(progress);
				success = cs.getSrc().copy(cs.getTgt());
				cs.setSuccess(success);

				// Update table element if action was successfully performed:
				if (success) {
					if (cs.isCopyFromSource()) {
						cs.getElement().setTgtFile(cs.getTgt());
					} else {
						cs.getElement().setSrcFile(cs.getTgt());
					}

					cs.getElement().revalidate();
				}
			}

			progress.decrease();
			i++;
		}

		// Update the progress object:
		progress.decrease();
		progress.setDuration(System.currentTimeMillis() - time);
		progress.finish();

		// Store the history, even if the synchronization process was cancelled:
		if (config.isStoreHistory() && !config.getTestRun()) {
			JFSHistory h = comparison.getHistory();
			h.update(comparison.getTable());
			h.store();
		}
	}

	/**
	 * Tests the algorithm.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		String srcDir = "test/dirs/source";
		String tgtDir = "test/dirs/target";
		boolean testrun = true;
		boolean testall = false;
		int syncMode = JFSSyncModes.getInstance().getDefaultMode();

		String usage = "Usage:\n" + "  JFSSynchronization [options]\n"
				+ "Options:\n" + "  -src <directory>  The source directory.\n"
				+ "  -tgt <directory>  The target directory.\n"
				+ "  -sync <mode>      Sets number for synchronization mode.\n"
				+ "  -notest           Attention: Turns off testrun.\n"
				+ "  -testall          Test all synchronization modes.";

		int i = 0;

		try {
			while (i < args.length) {
				if (args[i].equals("-sync")) {
					i++;

					byte newSyncMode = Byte.parseByte(args[i]);

					// Set only, if sync mode exists:
					if (JFSSyncModes.getInstance().contains(syncMode))
						syncMode = newSyncMode;

				} else if (args[i].equals("-notest")) {
					testrun = false;
				} else if (args[i].equals("-testall")) {
					testall = true;
				} else if (args[i].equals("-src")) {
					i++;
					srcDir = args[i];
				} else if (args[i].equals("-tgt")) {
					i++;
					tgtDir = args[i];
				} else {
					JFSLog.getOut().getStream().println(usage);
					System.exit(0);
				}

				i++;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
			JFSLog.getOut().getStream().println(usage);
			System.exit(0);
		}

		JFSProgress progress = new JFSProgress();
		JFSProgressPrint observer = new JFSProgressPrint();
		JFSDirectoryPair pair = new JFSDirectoryPair(srcDir, tgtDir);
		JFSSynchronization synchronization = new JFSSynchronization(pair,
				progress, new JFSQuestion());

		progress.attach(observer);
		synchronization.getComparison().compare();
		JFSPrint.printComparisonTable(synchronization.getComparison()
				.getTable());
		JFSConfig config = JFSConfig.getInstance();

		if (testall) {
			config.setTestRun(true);

			for (JFSSyncMode mode : JFSSyncModes.getInstance().getModes()) {
				config.setSyncMode((byte) mode.getId());
				synchronization.computeActions();
				synchronization.computeSynchronizationLists();
				synchronization.synchronize();
			}
		} else {
			config.setTestRun(testrun);
			config.setSyncMode((byte) syncMode);
			synchronization.computeActions();
			synchronization.computeSynchronizationLists();
			synchronization.synchronize();
		}
	}
}