/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync;

import java.io.File;
import java.util.HashMap;
import java.util.Vector;

import jfs.conf.JFSConfig;
import jfs.conf.JFSConfigObserver;
import jfs.conf.JFSDirectoryPair;
import jfs.conf.JFSLog;
import jfs.conf.JFSViewMode;
import jfs.conf.JFSViewModes;
import jfs.shell.JFSPrint;
import jfs.shell.JFSProgressPrint;

/**
 * The JFS task object performs comparisons and synchronizations for a set of
 * directory pairs specified in the configuration object. For this purpose, a
 * JFS synchronization object is used for each specified directory pair. The JFS
 * task is an observer of the configuration object (the user's profile). If the
 * configuration changes, corresponding updates are performed in order to
 * refelect the current configuration. Additionally, the JFS task creates
 * different views on the overall comparison table (containing the comparison
 * tables for all specified directory pairs. In order to perform a complete
 * synchronization cycle, the following actions have to be taken: (1) perform a
 * comparison for each directory pair, (2) compute the actions applied to the
 * comparison table according to the chosen synchronization mode, (3) compute
 * the views onto the comparison table, (4) compute the lists of copy and delete
 * statements from the actions, and finally (5) perform a synchronization.
 * 
 * @see #compare()
 * @see #computeActions()
 * @see #computeViews()
 * @see #computeSynchronizationLists()
 * @see #synchronize()
 * @author Jens Heidrich
 * @version $Id: JFSTask.java,v 1.26 2005/09/28 20:33:11 heidrich Exp $
 */
public class JFSTask implements JFSConfigObserver {

	/** Stores the only instance of the class. */
	private static JFSTask instance = null;

	/** Vector of JFS synchronization objects. */
	private Vector<JFSSynchronization> synchronizationList = new Vector<JFSSynchronization>();

	/** Monitors the algorithm's progress. */
	private JFSProgress progress = new JFSProgress();

	/** Answers questions during before the algorithm is performed. */
	private JFSQuestion question = new JFSQuestion();

	/** Map of available views on the overall comparison table. */
	private HashMap<Integer, Vector<JFSElement>> views = new HashMap<Integer, Vector<JFSElement>>();

	/** The current view onto the overall comparison table. */
	private Vector<JFSElement> currentView = null;

	/** The list of copy statements. */
	private Vector<JFSCopyStatement> copyStatements = new Vector<JFSCopyStatement>();

	/** The list of delete statements. */
	private Vector<JFSDeleteStatement> deleteStatements = new Vector<JFSDeleteStatement>();

	/**
	 * Creates a new task object.
	 */
	private JFSTask() {
		// Initialize map of views:
		for (JFSViewMode mode : JFSViewModes.getInstance().getModes()) {
			views.put(mode.getId(), new Vector<JFSElement>());
		}

		// Set default values:
		clean();
	}

	/**
	 * Returns the reference of the only instance.
	 * 
	 * @return The only instance.
	 */
	public final static JFSTask getInstance() {
		if (instance == null)
			instance = new JFSTask();

		return instance;
	}

	/**
	 * Restores the default values.
	 */
	private final void clean() {
		// Clear vectors:
		synchronizationList.clear();
		copyStatements.clear();
		deleteStatements.clear();

		// Clear all views:
		for (JFSViewMode mode : JFSViewModes.getInstance().getModes()) {
			views.get(mode.getId()).clear();
		}

		// Set default view:
		currentView = views.get(JFSViewModes.getInstance().getDefaultMode());
	}

	/**
	 * Returns the JFS progress object (e.g., in order to detach a new observer
	 * to the algorithm's progress).
	 * 
	 * @return The used JFS progress object.
	 */
	public final JFSProgress getProgress() {
		return progress;
	}

	/**
	 * Returns the JFS question object (e.g., in order to set the oracle).
	 * 
	 * @return The used JFS question object.
	 */
	public final JFSQuestion getQuestion() {
		return question;
	}

	/**
	 * @see JFSConfigObserver#updateConfig(JFSConfig)
	 */
	public final void updateConfig(JFSConfig config) {
		JFSViewMode mode = JFSViewModes.getInstance().get(config.getView());
		if (mode != null)
			currentView = views.get(mode.getId());
	}

	/**
	 * @see JFSConfigObserver#updateComparison(JFSConfig)
	 */
	public final void updateComparison(JFSConfig config) {
		// Clean the task object and producers:
		clean();
		JFSFileProducerManager.getInstance().resetProducers();

		// Fill the synchronization list:
		for (JFSDirectoryPair pair : config.getDirectoryList()) {
			// Create new synchronization object:
			JFSSynchronization synchronization = new JFSSynchronization(pair,
					progress, question);

			// Add to synchronization list:
			synchronizationList.add(synchronization);
		}

		// Update config:
		updateConfig(config);
	}

	/**
	 * @see JFSConfigObserver#updateServer(JFSConfig)
	 */
	public final void updateServer(JFSConfig config) {
		updateComparison(config);
	}

	/**
	 * Returns overall size of jfsTables. That is the sum of all table sizes
	 * within the aggregating vector jfsTables.
	 * 
	 * @return Overall size.
	 */
	public final long getOverallSize() {
		int size = 0;
		for (JFSSynchronization synchronization : synchronizationList) {
			size += synchronization.getComparison().getTable().size();
		}
		return size;
	}

	/**
	 * Returns the current view on the overall comparison table.
	 * 
	 * @return Vector of JFSElement objects.
	 */
	public final Vector<JFSElement> getViewTable() {
		return currentView;
	}

	/**
	 * @see JFSSynchronization#getCopyStatements()
	 */
	public final Vector<JFSCopyStatement> getCopyStatements() {
		return copyStatements;
	}

	/**
	 * @see JFSSynchronization#getDeleteStatements()
	 */
	public final Vector<JFSDeleteStatement> getDeleteStatements() {
		return deleteStatements;
	}

	/**
	 * Returns all failed copy statements.
	 * 
	 * @return Vector of all failed copy statements.
	 */
	public final Vector<JFSCopyStatement> getFailedCopyStatements() {
		Vector<JFSCopyStatement> v = new Vector<JFSCopyStatement>();

		for (JFSCopyStatement cs : copyStatements) {
			if (!cs.getSuccess() && cs.getCopyFlag())
				v.add(cs);
		}

		return v;
	}

	/**
	 * Returns all failed delete statements.
	 * 
	 * @return Vector of all failed delete statements.
	 */
	public final Vector<JFSDeleteStatement> getFailedDeleteStatements() {
		Vector<JFSDeleteStatement> v = new Vector<JFSDeleteStatement>();

		for (JFSDeleteStatement ds : deleteStatements) {
			if (!ds.getSuccess() && ds.getDeleteFlag())
				v.add(ds);
		}

		return v;
	}

	/**
	 * Starts comparison for all directory pairs, computes the actions that have
	 * to be taken according to the chosen synchronization mode.
	 * 
	 * @see JFSComparison#compare()
	 * @see #computeActions()
	 */
	public final void compare() {
		// Initialize the progress object:
		progress.start();
		progress.increase(synchronizationList.size(), 1);

		// Start comparison:
		for (JFSSynchronization synchronization : synchronizationList) {
			synchronization.getComparison().compare();
		}

		progress.end();
		progress.decrease();

		// Compute lists of files to copy and to delete:
		computeActions();
	}

	/**
	 * Computes the actions that have to be done for the overall comparison
	 * table according to the chosen synchronization mode and computes the views
	 * onto the comparison table accordingly.
	 * 
	 * @see JFSSynchronization#computeActions()
	 * @see JFSTask#computeViews()
	 */
	public final void computeActions() {
		// Compute actions for each synchronization object:
		for (JFSSynchronization synchronization : synchronizationList) {
			synchronization.computeActions();
		}

		// Compute the views on the overall comparison table:
		computeViews();
	}

	/**
	 * Compute all available views on the overall comparison table.
	 */
	private final void computeViews() {
		// Handle each specified view:
		for (JFSViewMode mode : JFSViewModes.getInstance().getModes()) {
			Vector<JFSElement> view = views.get(mode.getId());
			if (view == null)
				continue;
			view.clear();
			boolean viewed = false;

			// Process synchronization list in reverse order in order to
			// identify directories that contain files that have to
			// be copied (or deleted):
			for (int i = synchronizationList.size() - 1; i >= 0; i--) {
				JFSSynchronization synchronization = synchronizationList.get(i);

				// Process included vectors of JFSElement objects in reverse
				// order:
				Vector<JFSElement> elementList = synchronization
						.getComparison().getTable();

				for (int j = elementList.size() - 1; j >= 0; j--) {
					JFSElement te = elementList.get(j);

					if (mode.isViewed(te.getAction())) {
						view.add(te);
						viewed = true;
					} else if (te.isDirectory() && viewed) {
						// Insert in view directories that contain files that
						// were already viewed:
						view.add(te);
					}

					if (te.isDirectory())
						viewed = false;
				}
			}

			// Invert vectors of JFSElement objects in order to handle
			// all lists of JFSElement objects uniquely:
			JFSTask.invert(view);
		}
	}

	/**
	 * @see JFSSynchronization#computeSynchronizationLists()
	 */
	public final void computeSynchronizationLists() {
		// Clear the lists of files that have to copied and deleted:
		copyStatements.clear();
		deleteStatements.clear();

		// Compose overall lists out of single lists:
		for (JFSSynchronization synchronization : synchronizationList) {
			synchronization.computeSynchronizationLists();
			copyStatements.addAll(synchronization.getCopyStatements());
			deleteStatements.addAll(synchronization.getDeleteStatements());
		}
	}

	/**
	 * Performs the synchronization for all directory pairs, computes the
	 * updated actions afterwards, and shuts down the JFS file producers.
	 * 
	 * @see JFSSynchronization#synchronize()
	 * @see #computeActions()
	 * @see JFSComparison#shutDownProducers()
	 */
	public final void synchronize() {
		// Initialize the progress object:
		progress.start();
		progress.increase(synchronizationList.size(), 1);

		// Start synchronization:
		for (JFSSynchronization synchronization : synchronizationList) {
			synchronization.synchronize();
		}

		progress.end();
		progress.decrease();

		// Compute actions:
		computeActions();

		// Shuts down file producers of previously created comparison objects:
		for (JFSSynchronization s : synchronizationList) {
			s.getComparison().shutDownProducers();
		}
	}

	/**
	 * Inverts all elements within a vector; i.e., the first element will become
	 * the last one and so on.
	 * 
	 * @param <T>
	 *            The elements contained in the vector.
	 * @param v
	 *            The vector to be inverted.
	 */
	public static <T> void invert(Vector<T> v) {
		T temp;
		int size = v.size();

		for (int i = 0; i < size / 2; i++) {
			temp = v.get(i);
			v.set(i, v.get(size - 1 - i));
			v.set(size - 1 - i, temp);
		}
	}

	/**
	 * Tests the used algorithms.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		String configFile;

		if (args.length != 1) {
			configFile = "test/profiles/LocalSync.xml";
		} else {
			configFile = args[0];
		}

		// Setup objects:
		JFSConfig config = JFSConfig.getInstance();
		JFSTask task = JFSTask.getInstance();
		JFSProgress progress = task.getProgress();
		JFSProgressPrint observer = new JFSProgressPrint();

		progress.attach(observer);
		config.attach(task);

		// Load an existing configuation file:
		config.load(new File(configFile));
		config.simplePrint();

		JFSLog.getOut().getStream().println();
		task.compare();
		JFSLog.getOut().getStream().println();
		JFSPrint.printComparisonTable(task.getViewTable());
	}
}