/*
 * JFileSync
 * Copyright (C) 2002-2005, Jens Heidrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA, 02110-1301, USA
 */

package jfs.sync.zip;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import jfs.conf.JFSConfig;
import jfs.conf.JFSLog;
import jfs.sync.local.JFSLocalFile;

/**
 * Represents a zip file.
 * 
 * @author Jens Heidrich
 * @version $Id: JFSZipFile.java,v 1.5 2005/06/27 16:00:06 heidrich Exp $
 */
public class JFSZipFile extends JFSLocalFile {

	/**
	 * Creates a JFS file object out of an corresponding String that represents
	 * a file or directory name.
	 * 
	 * @param rootPath
	 *            The path of the assigned root JFS file.
	 * @param relativePath
	 *            The relative path of the JFS file starting from the root JFS
	 *            file.
	 */
	public JFSZipFile(String rootPath, String relativePath) {
		super(rootPath, relativePath);
	}

	/**
	 * Extracts the zip file to a new temporary directory.
	 * 
	 * @param zip
	 *            The zip file to extract.
	 * @return The created temporary directory.
	 */
	public static File start(String zip) {
		String tempDir = System.getProperty("java.io.tmpdir");
		File zipFile = new File(zip);
		File zipTempDir = new File(tempDir + File.separatorChar
				+ System.currentTimeMillis() + zipFile.getName());

		unzip(zip, zipTempDir.getPath());

		return zipTempDir;
	}

	/**
	 * Zips all files from a temporary directory and removes the directory
	 * afterwards. If a zip file of the same name already exists, it is
	 * overwritten.
	 * 
	 * @param zip
	 *            The name of the zip file to create.
	 * @param zipTempDir
	 *            The temporary directory.
	 */
	public void finish(String zip, File zipTempDir) {
		boolean success = zip(zip, zipTempDir.getPath());
		if (success) {
			success = success && deleteFile(zipTempDir);
		}
	}

	/**
	 * Traverses a file hierarchie for deleting all files.
	 * 
	 * @param file
	 *            The file to delete.
	 * @return True if everything went OK.
	 */
	private static boolean deleteFile(File file) {
		boolean success = true;

		if (file.isDirectory()) {
			for (File f : file.listFiles()) {
				success = success && deleteFile(f);
			}
		}

		success = success && file.delete();

		return success;
	}

	/**
	 * Unzips a zip file to a specified directory.
	 * 
	 * @param zip
	 *            Name of the zip file to extract.
	 * @param dir
	 *            Name of the directory to extract the zip file to.
	 * @return True if everything went OK.
	 */
	public static boolean unzip(String zip, String dir) {
		FileInputStream f = null;
		try {
			f = new FileInputStream(zip);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return false;
		}
		ZipInputStream zipIn = new ZipInputStream(f);
		ZipEntry entry = null;
		try {
			entry = zipIn.getNextEntry();
			while (entry != null) {
				String path = dir + File.separator + entry.getName();
				File entryFile = new File(path);

				if (entry.isDirectory()) {
					entryFile.mkdirs();
				} else {
					FileOutputStream entryOut = new FileOutputStream(entryFile);

					byte[] buf = new byte[JFSConfig.getInstance()
							.getBufferSize()];
					int len;

					while ((len = zipIn.read(buf)) > 0) {
						entryOut.write(buf, 0, len);
					}
					entryOut.close();
					entryFile.setLastModified(entry.getTime());
				}

				zipIn.closeEntry();
				entry = zipIn.getNextEntry();
			}

			zipIn.close();
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * Creates a zip file from a specified directory.
	 * 
	 * @param zip
	 *            Name of the zip file to create.
	 * @param dir
	 *            Name of the directory to create the zip file from.
	 * @return True if everything went OK.
	 */
	public static boolean zip(String zip, String dir) {
		FileOutputStream f = null;
		try {
			f = new FileOutputStream(zip);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return false;
		}
		File zipDir = new File(dir);
		ZipOutputStream zipOut = new ZipOutputStream(f);
		boolean success = zipTraverse(zipDir, zipOut, "");
		try {
			zipOut.close();
		} catch (IOException e1) {
			return false;
		}

		return success;
	}

	/**
	 * Traverses a file hierarchie for creating a zip file.
	 * 
	 * @param dir
	 *            The directory to start with.
	 * @param out
	 *            The zip output stream to write the contents to.
	 * @param path
	 *            The relative path to store the file to.
	 * @return True if everything went OK.
	 */
	private static boolean zipTraverse(File dir, ZipOutputStream out,
			String path) {
		boolean success = true;
		for (File f : dir.listFiles()) {
			String entryPath = path + f.getName();
			try {
				if (f.isDirectory()) {
					ZipEntry entry = new ZipEntry(entryPath + "/");
					out.putNextEntry(entry);
					out.closeEntry();
					success = success && zipTraverse(f, out, entryPath + "/");
				} else {
					ZipEntry entry = new ZipEntry(entryPath);
					entry.setTime(f.lastModified());
					entry.setMethod(ZipEntry.DEFLATED);
					out.putNextEntry(entry);
					FileInputStream entryIn = new FileInputStream(f);

					byte[] buf = new byte[JFSConfig.getInstance()
							.getBufferSize()];
					int len;

					while ((len = entryIn.read(buf)) > 0) {
						out.write(buf, 0, len);
					}
					entryIn.close();
					out.closeEntry();
				}
			} catch (Exception e) {
				return false;
			}
		}
		return success;
	}

	/**
	 * Tests the algorithm.
	 * 
	 * @param args
	 *            Command line arguments.
	 */
	public static void main(String[] args) {
		if (args.length != 3
				|| (!args[0].equals("-zip") && !args[0].equals("-unzip"))) {
			JFSLog.getOut().getStream().println(
					"Usage: [-zip|-unzip] [Zip File] [Directory]");
		} else {
			if (args[0].equals("-zip")) {
				zip(args[1], args[2]);
			} else if (args[0].equals("-unzip")) {
				unzip(args[1], args[2]);
			}
		}
	}
}